from enum import Enum, EnumMeta
from typing import List

class StrEnumMeta(EnumMeta):
    # this is workaround for submitit pickling leading to instance checks failing in hydra for StrEnum, see
    # https://github.com/facebookresearch/hydra/issues/1156
    @classmethod
    def __instancecheck__(cls, other):
        return "enum" in str(type(other))


class StrEnum(Enum, metaclass=StrEnumMeta):
    def __str__(self):
        return self.value

    def __eq__(self, other: str):
        return self.value == other

    def __repr__(self):
        return self.value

    def __hash__(self):
        return hash(str(self))

def ChoiceEnum(choices: List[str]):
    """return the Enum class used to enforce list of choices"""
    return StrEnum("Choices", {k: k for k in choices})

LOG_FORMAT_CHOICES = ChoiceEnum(["json", "none", "simple", "tqdm", "csv"])


from dataclasses import is_dataclass
from omegaconf import OmegaConf, open_dict

from genhpf.configs import BaseConfig

def merge_with_parent(dc: BaseConfig, cfg: BaseConfig, remove_missing=False):
    if remove_missing:

        def remove_missing_rec(src_keys, target_cfg):
            if is_dataclass(target_cfg):
                target_keys = set(target_cfg.__dataclass_fields__.keys())
            else:
                target_keys = set(target_cfg.keys())
            
            for k in list(src_keys.keys()):
                if k not in target_keys:
                    del src_keys[k]
                elif OmegaConf.is_config(src_keys[k]):
                    tgt = getattr(target_cfg, k)
                    if tgt is not None and (is_dataclass(tgt) or hasattr(tgt, "keys")):
                        remove_missing_rec(src_keys[k], tgt)
    
        with open_dict(cfg):
            remove_missing_rec(cfg, dc)

    merged_cfg = OmegaConf.merge(dc, cfg)
    merged_cfg.__dict__["_parent"] = cfg.__dict__["_parent"]
    OmegaConf.set_struct(merged_cfg, True)
    return merged_cfg


import logging
import time
from itertools import chain
from typing import Any, Dict, List, Tuple

import torch
import torch.distributed as dist
import torch.utils
import torch.utils.data
from omegaconf import OmegaConf
from torch.optim import Adam

from genhpf.configs import Config
from genhpf.datasets import BaseDataset
from genhpf.loggings import metrics
from genhpf.utils import checkpoint_utils, distributed_utils, utils
from genhpf.utils.file_io import PathManager

logger = logging.getLogger(__name__)


class Trainer(object):
    def __init__(self, cfg: Config, model, criterion):
        self.cfg = cfg
        self.cuda = torch.cuda.is_available()
        if self.cuda:
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")

        self._criterion = criterion
        self._model = model

        self._criterion = self._criterion.to(device=self.device)
        self._model = self._model.to(device=self.device)

        self._num_updates = 0

        self._optimizer = None
        self._wrapped_criterion = None
        self._wrapped_model = None

        if self.cuda:
            self.cuda_env = utils.CudaEnvironment()
            if self.data_parallel_world_size > 1:
                self.cuda_env_arr = distributed_utils.all_gather_list(
                    self.cuda_env, group=distributed_utils.get_global_group()
                )
            else:
                self.cuda_env_arr = [self.cuda_env]
            if self.data_parallel_rank == 0:
                utils.CudaEnvironment.pretty_print_cuda_env_list(self.cuda_env_arr)
        else:
            self.cuda_env = None
            self.cuda_env_arr = None

        metrics.log_start_time("wall", priority=790, round=0)

        self._start_time = time.time()
        self._previous_training_time = 0
        self._cumulative_training_time = None

    @property
    def data_parallel_world_size(self):
        if self.cfg.distributed_training.distributed_world_size == 1:
            return 1
        return distributed_utils.get_data_parallel_world_size()

    @property
    def data_parallel_process_group(self):
        return distributed_utils.get_data_parallel_group()

    @property
    def data_parallel_rank(self):
        if self.cfg.distributed_training.distributed_world_size == 1:
            return 0
        return distributed_utils.get_data_parallel_rank()

    @property
    def is_data_parallel_master(self):
        return self.data_parallel_rank == 0

    @property
    def use_distributed_wrapper(self) -> bool:
        return self.data_parallel_world_size > 1

    @property
    def criterion(self):
        if self._wrapped_criterion is None:
            if utils.has_parameters(self._criterion) and self.use_distributed_wrapper:
                self._wrapped_criterion = distributed_utils.DistributedModel(
                    self.cfg.distributed_training,
                    self._criterion,
                    process_group=self.data_parallel_process_group,
                    device=self.device,
                )
            else:
                self._wrapped_criterion = self._criterion
        return self._wrapped_criterion

    @property
    def model(self):
        if self._wrapped_model is None:
            if self.use_distributed_wrapper:
                self._wrapped_model = distributed_utils.DistributedModel(
                    self.cfg.distributed_training,
                    self._model,
                    process_group=self.data_parallel_process_group,
                    device=self.device,
                )
            else:
                self._wrapped_model = self._model
        return self._wrapped_model

    @property
    def optimizer(self):
        if self._optimizer is None:
            self._build_optimizer()
        return self._optimizer

    def _build_optimizer(self):
        params = list(
            filter(lambda p: p.requires_grad, chain(self.model.parameters(), self.criterion.parameters()))
        )
        self._optimizer = Adam(
            params,
            lr=self.cfg.optimization.lr,
            betas=self.cfg.optimization.adam_betas,
            eps=self.cfg.optimization.adam_eps,
            weight_decay=self.cfg.optimization.weight_decay,
        )

    def state_dict(self):
        pretraining_parameter_names = []
        if hasattr(self.model, "get_pretraining_parameter_names"):
            pretraining_parameter_names = self.model.get_pretraining_parameter_names()
        model_state_dict = {
            k: v for k, v in self.model.state_dict().items() if k not in pretraining_parameter_names
        }

        state_dict = {
            "cfg": (
                OmegaConf.to_container(self.cfg, resolve=True, enum_to_str=True)
                if OmegaConf.is_config(self.cfg)
                else self.cfg
            ),
            "model": model_state_dict,
        }

        return state_dict

    def save_checkpoint(self, filename):
        """Save all training state in a checkpoint file."""
        logger.info(f"Saving checkpoint to {filename}")

        state_dict = utils.move_to_cpu(self.state_dict())
        if self.is_data_parallel_master:
            checkpoint_utils.torch_persistent_save(state_dict, filename, async_write=False)
        logger.info(f"Finished saving checkpoint to {filename}")

    def load_checkpoint(self, filename) -> None:
        """
        Load all training state from a checkpoint file.
        rank = 0 will load the checkpoint and broadcast it to other ranks.
        """
        logger.info(f"Loading checkpoint from {filename}")
        is_distributed = self.data_parallel_world_size > 1
        bexists = PathManager.isfile(filename)
        if bexists:
            if self.data_parallel_rank == 0:
                state = checkpoint_utils.load_checkpoint_to_cpu(filename)
            else:
                state = None

            if is_distributed:
                state = distributed_utils.broadcast_object(
                    state, src_rank=0, group=self.data_parallel_process_group, dist_device=self.device
                )

            # load model parameters
            try:
                self.model.load_state_dict(state["model"], strict=True)
                # save memory for later steps
                del state["model"]
            except Exception:
                raise Exception(
                    f"Cannot load model parameters from checkpoint {filename}; "
                    "please ensure that the architectures match."
                )
        else:
            logger.info(f"No existing checkpoint found {filename}")

    def get_train_iterator(
        self,
        dataset: BaseDataset,
    ) -> Tuple[torch.utils.data.DataLoader, torch.utils.data.Sampler]:
        """Return an DataLoader instance for training."""
        batch_sampler = (
            torch.utils.data.DistributedSampler(dataset, shuffle=True) if dist.is_initialized() else None
        )
        batch_iterator = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.cfg.dataset.batch_size,
            shuffle=True if not dist.is_initialized() else False,
            num_workers=self.cfg.dataset.num_workers if not self.cfg.common.debug else 0,
            collate_fn=dataset.collator,
            sampler=batch_sampler,
        )
        return batch_iterator, batch_sampler

    def get_valid_iterator(
        self,
        dataset: BaseDataset,
    ) -> Tuple[torch.utils.data.DataLoader, torch.utils.data.Sampler]:
        """Return an DataLoader instance for validation."""
        batch_sampler = (
            torch.utils.data.DistributedSampler(dataset, shuffle=False) if dist.is_initialized() else None
        )
        batch_iterator = torch.utils.data.DataLoader(
            dataset,
            batch_size=self.cfg.dataset.batch_size,
            shuffle=False,
            num_workers=self.cfg.dataset.num_workers if not self.cfg.common.debug else 0,
            collate_fn=dataset.collator,
            sampler=batch_sampler,
        )
        return batch_iterator, batch_sampler

    @metrics.aggregate("train")
    def train_step(self, sample):
        """Do forward, backward and optimization steps."""

        self._set_seed()
        self.model.train()
        self.criterion.train()
        self.zero_grad()

        metrics.log_start_time("train_wall", priority=800, round=0)

        logging_outputs = []
        sample = utils.prepare_sample(sample)

        loss, sample_size, logging_output = self.criterion(self.model, sample)
        if loss.item() > 0:
            loss.backward()  # backward
            self.optimizer.step()

        logging_outputs.append(logging_output)

        # emptying the CUDA cache after the first step can
        # reduce the chance of OOM
        if self.cuda and self.get_num_updates() == 0:
            torch.cuda.empty_cache()

        sample_size = float(sample_size)

        if self._sync_stats():
            train_time = self._local_cumulative_training_time()

            logging_outputs, (sample_size, total_train_time) = self._aggregate_logging_outputs(
                logging_outputs, sample_size, train_time
            )

            self._cumulative_training_time = total_train_time / self.data_parallel_world_size

        logging_output = None
        self.set_num_updates(self.get_num_updates() + 1)

        if self.cuda and self.cuda_env is not None:
            gb_used = torch.cuda.max_memory_allocated() / 1024 / 1024 / 1024
            torch.cuda.reset_peak_memory_stats()
            gb_free = self.cuda_env.total_memory_in_GB - gb_used
            metrics.log_scalar("gb_free", gb_free, priority=1500, round=1, weight=0)

        # extract private logs (usually only for valid steps) before logging
        logging_outputs = list(
            map(lambda x: {key: x[key] for key in x if not key.startswith("_")}, logging_outputs)
        )
        # log stats
        logging_output = self._reduce_and_log_stats(logging_outputs, sample_size)

        metrics.log_stop_time("train_wall")
        return logging_output

    @metrics.aggregate("valid")
    def valid_step(self, sample, subset=None):
        """Do forward pass in evaluation mode."""
        with torch.no_grad():
            self.model.eval()
            self.criterion.eval()

            sample = utils.prepare_sample(sample)
            _loss, sample_size, logging_output = self.criterion(self.model, sample)
            logging_outputs = [logging_output]

        # gather logging outputs from all replicas
        if self.data_parallel_world_size > 1:
            logging_outputs, (sample_size,) = self._aggregate_logging_outputs(
                logging_outputs,
                sample_size,
                ignore=False,
            )

        # log validation stats
        logging_output = self._reduce_and_log_stats(logging_outputs, sample_size)

        return _loss, sample_size, logging_outputs

    def zero_grad(self):
        self.optimizer.zero_grad()

    def get_num_updates(self):
        return self._num_updates

    def set_num_updates(self, num_updates):
        self._num_updates = num_updates
        metrics.log_scalar("num_updates", self._num_updates, weight=0, priority=200)

    def cumulative_training_time(self):
        if self._cumulative_training_time is None:
            return self._local_cumulative_training_time()
        else:
            return self._cumulative_training_time

    def _local_cumulative_training_time(self):
        return time.time() - self._start_time + self._previous_training_time

    def _set_seed(self):
        seed = self.cfg.common.seed + self.get_num_updates()
        utils.set_torch_seed(seed)

    def _sync_stats(self):
        if self.data_parallel_world_size == 1:
            return False
        else:
            return True

    def _aggregate_logging_outputs(
        self, logging_outputs: List[Dict[str, Any]], *extra_stats_to_sum, ignore=False
    ):
        return self._all_gather_list_sync(logging_outputs, *extra_stats_to_sum, ignore=ignore)

    def _all_gather_list_sync(self, logging_outputs: List[Dict[str, Any]], *extra_stats_to_sum, ignore=False):
        if ignore:
            logging_outputs = []
        results = list(
            zip(
                *distributed_utils.all_gather_list(
                    [logging_outputs] + list(extra_stats_to_sum),
                    max_size=getattr(self.cfg.common, "all_gather_list_size", 32768),
                    group=self.data_parallel_process_group,
                )
            )
        )
        logging_outputs, extra_stats_to_sum = results[0], results[1:]
        logging_outputs = list(chain.from_iterable(logging_outputs))
        extra_stats_to_sum = [sum(s) for s in extra_stats_to_sum]
        return logging_outputs, extra_stats_to_sum

    def _reduce_and_log_stats(self, logging_outputs, sample_size):
        metrics.log_speed("ups", 1.0, priority=100, round=2)

        with metrics.aggregate() as agg:
            if logging_outputs is not None:
                self.criterion.__class__.reduce_metrics(logging_outputs)
                del logging_outputs

            logging_output = agg.get_smoothed_values()
            logging_output["sample_size"] = sample_size
            return logging_output



import logging
from typing import Union

import numpy as np
import torch.utils.data

from genhpf.configs import ChoiceEnum

MASK_UNIT_CHOICES = ChoiceEnum(["token", "event"])

logger = logging.getLogger(__name__)


class BaseDataset(torch.utils.data.Dataset):
    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    def collator(self, samples):
        raise NotImplementedError

    def mask(
        self,
        tokens: Union[np.ndarray, torch.Tensor],
        mask_prob: float,
        vocab_size: int,
        mask_token_id: int,
        mask_unit: MASK_UNIT_CHOICES = "token",
        sep_token_id: int = None,
    ):
        """
        Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original.
        """
        assert 0.0 < mask_prob < 1.0, "mask_prob must be in the range (0.0, 1.0)"

        if isinstance(tokens, np.ndarray):
            tokens = torch.LongTensor(tokens)
        tokens = tokens.long()
        labels = tokens.clone()

        assert tokens.dim() == 2, (
            "input tokens must be 2D tensor, where the first dimension is the number of embeddings "
            "(i.e., input_ids, type_ids, dpe_ids) and the second dimension is the length of token "
            "sequence."
        )

        if mask_unit == "token":
            probability_matrix = torch.full((labels.size(-1),), mask_prob)
            # do not mask special tokens
            if not hasattr(self, "tokenizer"):
                from transformers import AutoTokenizer

                self.tokenizer = AutoTokenizer.from_pretrained("emilyalsentzer/Bio_ClinicalBERT")

            special_tokens_mask = torch.tensor(
                self.tokenizer.get_special_tokens_mask(labels[0], already_has_special_tokens=True),
                dtype=torch.bool,
            )
            probability_matrix.masked_fill_(special_tokens_mask, value=0.0)
            probability_matrix[torch.where(labels[0] == self.pad_token_id)] = 0.0

            mask_indices = torch.bernoulli(probability_matrix).bool()
            while mask_indices.sum() == 0:
                mask_indices = torch.bernoulli(probability_matrix).bool()
        elif mask_unit == "event":
            if sep_token_id is None:
                logger.warning(
                    "sep_token_id is not provided. Using the default [SEP] token id (102) as "
                    "the event delimiter."
                )
                sep_token_id = 102  # token id for [SEP]
            event_indices = torch.where(tokens[0] == sep_token_id)[0]
            assert len(event_indices) > 1, (
                "there must be at least two events in the input sequence to apply span masking. "
                "check if you are using the hierarchical structure which is not supporting the "
                "span masking method."
            )
            mask_indices = torch.zeros_like(tokens[0]).bool()
            masked_event_indices = torch.randperm(len(event_indices) - 1)[
                : round((len(event_indices) - 1) * mask_prob)
            ]
            for i in masked_event_indices:
                mask_indices[event_indices[i] : event_indices[i + 1]] = True
        else:
            raise ValueError(f"mask_unit must be one of {MASK_UNIT_CHOICES}")

        labels[:, ~mask_indices] = self.ignore_index  # we only compute loss on masked tokens

        # 80% of the time, we replace the masked input tokens with tokenizer.mask_token ([MASK])
        mask_positions = torch.bernoulli(torch.full(mask_indices.shape, 0.8)).bool() & mask_indices
        tokens[0, mask_positions] = mask_token_id  # for input_ids
        tokens[1, mask_positions] = 4  # for type_ids
        tokens[2, mask_positions] = 15  # for dpe_ids

        # 10% of the time, we replace the masked input tokens with random word
        random_positions = (
            torch.bernoulli(torch.full(mask_indices.shape, 0.5)).bool() & mask_indices & ~mask_positions
        )
        random_words = torch.randint(vocab_size, mask_indices.shape, dtype=torch.long)
        tokens[0, random_positions] = random_words[random_positions]
        random_types = torch.randint(7, mask_indices.shape, dtype=torch.long)
        tokens[1, random_positions] = random_types[random_positions]
        random_dpes = torch.randint(16, mask_indices.shape, dtype=torch.long)
        tokens[2, random_positions] = random_dpes[random_positions]

        # the rest of the time, we keep the masked input tokens unchanged
        return tokens, labels



"""
Wrapper around various loggers and progress bars (e.g., tqdm).
"""

import json
import logging
import os
import sys
from collections import OrderedDict
from contextlib import contextmanager
from numbers import Number
from typing import Optional

import torch

from .meters import AverageMeter, StopwatchMeter, TimeMeter

from genhpf.utils import distributed_utils as dist_utils

logger = logging.getLogger(__name__)


def progress_bar(
    iterator,
    log_format: Optional[str] = None,
    log_interval: int = 100,
    log_file: Optional[str] = None,
    epoch: Optional[int] = None,
    prefix: Optional[str] = None,
    default_log_format: str = "tqdm",
    wandb_project: Optional[str] = None,
    wandb_entity: Optional[str] = None,
    wandb_run_name: Optional[str] = None,
):
    if log_format is None:
        log_format = default_log_format
    if log_file is not None:
        handler = logging.FileHandler(filename=log_file)
        logger.addHandler(handler)

    if log_format == "tqdm" and not sys.stderr.isatty():
        log_format = "simple"

    if log_format == "json":
        bar = JsonProgressBar(iterator, epoch, prefix, log_interval)
    elif log_format == "none":
        bar = NoopProgressBar(iterator, epoch, prefix)
    elif log_format == "simple":
        bar = SimpleProgressBar(iterator, epoch, prefix, log_interval)
    elif log_format == "tqdm":
        bar = TqdmProgressBar(iterator, epoch, prefix)
    elif log_format == "csv":
        bar = CsvProgressBar(iterator, epoch, prefix, log_interval)
    else:
        raise ValueError("Unknown log format: {}".format(log_format))

    if wandb_project:
        bar = WandBProgressBarWrapper(bar, wandb_project, wandb_entity, run_name=wandb_run_name)

    return bar


def build_progress_bar(
    args,
    iterator,
    epoch: Optional[int] = None,
    prefix: Optional[str] = None,
    default: str = "tqdm",
    no_progress_bar: str = "none",
):
    """Legacy wrapper that takes an argparse.Namespace."""
    if getattr(args, "no_progress_bar", False):
        default = no_progress_bar
    tensorboard_logdir = None
    return progress_bar(
        iterator,
        log_format=args.log_format,
        log_interval=args.log_interval,
        epoch=epoch,
        prefix=prefix,
        tensorboard_logdir=tensorboard_logdir,
        default_log_format=default,
    )


def format_stat(stat):
    if isinstance(stat, Number):
        stat = "{:g}".format(stat)
    elif isinstance(stat, AverageMeter):
        stat = "{:.3f}".format(stat.avg)
    elif isinstance(stat, TimeMeter):
        stat = "{:g}".format(round(stat.avg))
    elif isinstance(stat, StopwatchMeter):
        stat = "{:g}".format(round(stat.sum))
    elif torch.is_tensor(stat):
        stat = stat.tolist()
    return stat


class BaseProgressBar(object):
    """Abstract class for progress bars."""

    def __init__(self, iterable, epoch=None, prefix=None):
        self.iterable = iterable
        self.n = getattr(iterable, "n", 0)
        self.epoch = epoch
        self.prefix = ""
        if epoch is not None:
            self.prefix += "epoch {:03d}".format(epoch)
        if prefix is not None:
            self.prefix += (" | " if self.prefix != "" else "") + prefix

    def __len__(self):
        return len(self.iterable)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def __iter__(self):
        raise NotImplementedError

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        raise NotImplementedError

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        raise NotImplementedError

    def update_config(self, config):
        """Log latest configuration."""
        pass

    def _str_commas(self, stats):
        return ", ".join(key + "=" + stats[key].strip() for key in stats.keys())

    def _str_pipes(self, stats):
        return " | ".join(key + " " + stats[key].strip() for key in stats.keys())

    def _format_stats(self, stats):
        postfix = OrderedDict(stats)
        # Preprocess stats according to datatype
        for key in postfix.keys():
            postfix[key] = str(format_stat(postfix[key]))
        return postfix


@contextmanager
def rename_logger(logger, new_name):
    old_name = logger.name
    if new_name is not None:
        logger.name = new_name
    yield logger
    logger.name = old_name


class JsonProgressBar(BaseProgressBar):
    """Log output in JSON format."""

    def __init__(self, iterable, epoch=None, prefix=None, log_interval=1000):
        super().__init__(iterable, epoch, prefix)
        self.log_interval = log_interval
        self.i = None
        self.size = None

    def __iter__(self):
        self.size = len(self.iterable)
        for i, obj in enumerate(self.iterable, start=self.n):
            self.i = i
            yield obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        step = step or self.i or 0

        if step > 0 and self.log_interval is not None and step % self.log_interval == 0:
            update = (
                self.epoch - 1 + (self.i + 1) / float(self.size)
                if self.epoch is not None
                else None
            )
            stats = self._format_stats(stats, epoch=self.epoch, update=update)
            with rename_logger(logger, tag):
                logger.info(json.dumps(stats))

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        self.stats = stats
        if tag is not None:
            self.stats = OrderedDict(
                [(tag + "_" + k, v) for k, v in self.stats.items()]
            )
        stats = self._format_stats(self.stats, epoch=self.epoch)
        with rename_logger(logger, tag):
            logger.info(json.dumps(stats))

    def _format_stats(self, stats, epoch=None, update=None):
        postfix = OrderedDict()
        if epoch is not None:
            postfix["epoch"] = epoch
        if update is not None:
            postfix["update"] = round(update, 3)
        # Preprocess stats according to datatype
        for key in stats.keys():
            if not key.startswith("_"):
                postfix[key] = format_stat(stats[key])
        return postfix

class NoopProgressBar(BaseProgressBar):
    """No logging."""

    def __init__(self, iterable, epoch=None, prefix=None):
        super().__init__(iterable, epoch, prefix)

    def __iter__(self):
        for obj in self.iterable:
            yield obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        pass

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        pass


class SimpleProgressBar(BaseProgressBar):
    """A minimal logger for non-TTY environments."""

    def __init__(self, iterable, epoch=None, prefix=None, log_interval=1000):
        super().__init__(iterable, epoch, prefix)
        self.log_interval = log_interval
        self.i = None
        self.size = None

    def __iter__(self):
        self.size = len(self.iterable)
        for i, obj in enumerate(self.iterable, start=self.n):
            self.i = i
            yield obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        step = step or self.i or 0
        if step > 0 and self.log_interval is not None and step % self.log_interval == 0:
            stats = self._format_stats(stats)
            postfix = self._str_commas(stats)
            with rename_logger(logger, tag):
                logger.info(
                    "{}:  {:5d} / {:d} {}".format(
                        self.prefix, self.i + 1, self.size, postfix
                    )
                )

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        postfix = self._str_pipes(self._format_stats(stats))
        with rename_logger(logger, tag):
            logger.info("{} | {}".format(self.prefix, postfix))


class TqdmProgressBar(BaseProgressBar):
    """Log to tqdm."""

    def __init__(self, iterable, epoch=None, prefix=None):
        super().__init__(iterable, epoch, prefix)
        from tqdm import tqdm

        self.tqdm = tqdm(
            iterable,
            self.prefix,
            leave=False,
            disable=(logger.getEffectiveLevel() > logging.INFO),
        )

    def __iter__(self):
        return iter(self.tqdm)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        self.tqdm.set_postfix(self._format_stats(stats), refresh=False)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        postfix = self._str_pipes(self._format_stats(stats))
        with rename_logger(logger, tag):
            logger.info("{} | {}".format(self.prefix, postfix))

try:
    import csv
except ImportError:
    csv = None

class CsvProgressBar(BaseProgressBar):
    """Log to csv."""

    def __init__(self, iterable, epoch=None, prefix=None, log_interval=1000):
        super().__init__(iterable, epoch, prefix)
        self.log_interval = log_interval
        self.i = None
        self.size = None

    def __iter__(self):
        self.size = len(self.iterable)
        for i, obj in enumerate(self.iterable, start=self.n):
            self.i = i
            yield obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to csv."""
        self._log_to_csv(stats, tag, step)
        self._print_log(stats, tag, step)

    def print(self, stats, tag=None, step=None):
        """Log end-of-epoch stats."""
        self._log_to_csv(stats, tag, step)
        self._print_log(stats, tag, step)

    def _print_log(self, stats, tag=None, step=None):
        """Print intermediate stats."""
        step = step or self.i or 0
        
        if step > 0 and self.log_interval is not None and step % self.log_interval == 0:
            stats = self._format_stats(stats)
            postfix = self._str_commas(stats)
            with rename_logger(logger, tag):
                logger.info(
                    "{}:  {:5d} / {:d} {}".format(
                        self.prefix, self.i + 1, self.size, postfix
                    )
                )

    def _log_to_csv(self, stats, tag=None, step=None):
        if csv is None:
            return
        if dist_utils.get_data_parallel_world_size() > 1 and dist_utils.get_data_parallel_rank() > 0:
            return

        csv_logs = {}

        if step is None:
            csv_logs["step"] = stats["num_updates"] if "num_updates" in stats else None
        else:
            csv_logs["step"] = step

        for key in stats.keys() - {"num_updates"}:
            if isinstance(stats[key], AverageMeter):
                csv_logs[key] = stats[key].val
            elif isinstance(stats[key], Number):
                csv_logs[key] = stats[key]

        fname = "log.csv" if tag is None else tag + ".csv"
        if not os.path.exists(fname):
            cols = ["step"]
            for key in csv_logs.keys() - {"step"}:
                cols.append(key)
            with open(fname, "w") as f:
                wr = csv.writer(f)
                wr.writerow(cols)
        
        with open(fname, "r") as f:
            rd = csv.reader(f)
            headers = next(iter(rd))

        appended_lth = 0
        for key in csv_logs.keys():
            if key not in headers:
                headers.append(key)
                appended_lth += 1
        
        if appended_lth > 0:
            lines = [headers]
            with open(fname, "r") as f:
                rd = csv.reader(f)
                # drop headers
                next(iter(rd))
                for line in rd:
                    for _ in range(appended_lth):
                        line.append("")
                    lines.append(line)
            with open(fname, "w") as f:
                wr = csv.writer(f)
                wr.writerows(lines)

        with open(fname, "a") as f:
            wr = csv.DictWriter(f, fieldnames=headers)
            wr.writerow(csv_logs)

try:
    import wandb
except ImportError:
    wandb = None

class WandBProgressBarWrapper(BaseProgressBar):
    """Log to Weights & Biases."""

    def __init__(self, wrapped_bar, wandb_project, wandb_entity, run_name=None):
        self.wrapped_bar = wrapped_bar
        if wandb is None:
            logger.warning("wandb not found, pip install wandb")
            return

        # reinit=False to ensure if wandb.init() is called multiple times
        # within one process it still references the same run
        wandb.init(project=wandb_project, entity=wandb_entity, reinit=False, name=run_name)

    def __iter__(self):
        return iter(self.wrapped_bar)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats to wandb."""
        self._log_to_wandb(stats, tag, step)
        self.wrapped_bar.log(stats, tag=tag, step=step)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        self._log_to_wandb(stats, tag, step)
        self.wrapped_bar.print(stats, tag=tag, step=step)

    def update_config(self, config):
        """Log latest configuration."""
        if wandb is not None:
            wandb.config.update(config)
        self.wrapped_bar.update_config(config)

    def _log_to_wandb(self, stats, tag=None, step=None):
        if wandb is None:
            return
        if step is None:
            step = stats["num_updates"] if "num_updates" in stats else None

        prefix = "" if tag is None else tag + "/"

        wandb_logs = {}
        for key in stats.keys() - {"num_updates"}:
            if isinstance(stats[key], AverageMeter):
                wandb_logs[prefix + key] = stats[key].val
            elif isinstance(stats[key], Number):
                wandb_logs[prefix + key] = stats[key]

        wandb.log(wandb_logs, step=step)


import logging
from typing import List, Union

import h5pickle
import numpy as np
import pandas as pd
import torch

from genhpf.datasets.dataset import BaseDataset

logger = logging.getLogger(__name__)


class GenHPFDataset(BaseDataset):
    def __init__(
        self,
        manifest_paths: List[str],
        structure: str,
        vocab_size: int = 28996,
        pad_token_id: int = 0,
        sep_token_id: int = 102,
        ignore_index: int = -100,
        apply_mask: bool = False,
        mask_token_id: int = 103,
        mask_prob: float = 0,
        mask_unit: str = "individual",
        simclr: bool = False,
        **kwargs,
    ):
        super().__init__()

        if structure == "hierarchical":
            structure = "hi"
        elif structure == "flattened":
            structure = "fl"
        self.structure = structure

        self.pad_token_id = pad_token_id

        self.ignore_index = ignore_index

        self.apply_mask = apply_mask
        self.mask_prob = mask_prob
        self.vocab_size = vocab_size
        self.mask_token_id = mask_token_id
        self.mask_unit = mask_unit
        self.sep_token_id = sep_token_id

        self.simclr = simclr

        for k, v in kwargs.items():
            self.__setattr__(k, v)

        self.data = []
        self.subjects = []
        self.labels = {}
        for i, manifest_path in enumerate(manifest_paths):
            with open(manifest_path, "r") as f:
                data_root = f.readline().strip()
                label_root = f.readline().strip()
                self.data.append(h5pickle.File(data_root, "r")["ehr"])
                labels = pd.read_csv(label_root)
                labels.index = [(i, str(x)) for x in labels["stay_id"]]
                labels = labels.drop(columns=["stay_id"])
                self.labels.update(labels.to_dict(orient="index"))
                for line in f:
                    items = line.strip().split("\t")
                    assert len(items) == 1, line
                    self.subjects.append((i, items[0]))
        logger.info(f"loaded {len(self.subjects)} samples from {len(manifest_paths)} dataset(s)")

    def __len__(self):
        return len(self.subjects)

    def __getitem__(self, index):
        raise NotImplementedError


class HierarchicalGenHPFDataset(GenHPFDataset):
    def __init__(
        self,
        manifest_paths: List[str],
        label: bool = False,
        tasks: List[str] = None,
        num_labels: List[int] = None,
        dummy_token_id: int = 101,
        **kwargs,
    ):
        kwargs.pop("structure", None)
        super().__init__(manifest_paths=manifest_paths, structure="hierarchical", **kwargs)

        self.label = label
        self.tasks = tasks
        self.num_labels = num_labels
        self.dummy_token_id = dummy_token_id

    def mask(self, tokens: Union[np.ndarray, torch.Tensor], **kwargs):
        for i, event in enumerate(tokens):
            tokens[i], _ = super().mask(event, **kwargs)
        return tokens

    def collator(self, samples):
        samples = [s for s in samples if s["input_ids"] is not None]
        if len(samples) == 0:
            return {}

        if self.simclr:
            input_ids = sum(
                [
                    [s["input_ids"][: len(s["input_ids"]) // 2], s["input_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
            type_ids = sum(
                [
                    [s["type_ids"][: len(s["input_ids"]) // 2], s["type_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
            dpe_ids = sum(
                [
                    [s["dpe_ids"][: len(s["input_ids"]) // 2], s["dpe_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
        else:
            input_ids = [s["input_ids"] for s in samples]
            type_ids = [s["type_ids"] for s in samples]
            dpe_ids = [s["dpe_ids"] for s in samples]

        sizes = [s.size(0) for s in input_ids]
        target_size = max(sizes)

        collated_input_ids = (
            input_ids[0].new_zeros((len(input_ids), target_size, len(input_ids[0][0]))).long()
        )
        collated_type_ids = type_ids[0].new_zeros((len(type_ids), target_size, len(type_ids[0][0]))).long()
        collated_dpe_ids = dpe_ids[0].new_zeros((len(dpe_ids), target_size, len(dpe_ids[0][0]))).long()
        for i, size in enumerate(sizes):
            diff = size - target_size
            if diff == 0:
                collated_input_ids[i] = input_ids[i]
                collated_type_ids[i] = type_ids[i]
                collated_dpe_ids[i] = dpe_ids[i]
            elif diff < 0:
                collated_input_ids[i] = torch.cat(
                    [
                        input_ids[i],
                        input_ids[i].new_zeros(-diff, len(input_ids[i][0])),
                    ],
                    dim=0,
                )
                # add dummy token to the start of each padded event as the event encoder can be
                # crushed when all the input tokens are pad tokens
                collated_input_ids[i][diff:, 0] = self.dummy_token_id
                collated_type_ids[i] = torch.cat(
                    [
                        type_ids[i],
                        type_ids[i].new_zeros(-diff, len(type_ids[i][0])),
                    ],
                    dim=0,
                )
                collated_dpe_ids[i] = torch.cat(
                    [
                        dpe_ids[i],
                        dpe_ids[i].new_zeros(-diff, len(dpe_ids[i][0])),
                    ],
                    dim=0,
                )
            else:
                raise ValueError(f"size mismatch, expected <={target_size}, got {size}")

        out = {"id": [s["id"] for s in samples]}
        out["net_input"] = {
            "input_ids": collated_input_ids,
            "type_ids": collated_type_ids,
            "dpe_ids": collated_dpe_ids,
        }

        if self.label:
            label = {}
            for task in self.tasks:
                label[task] = torch.stack([s[task] for s in samples])
            out["label"] = label

        return out

    def __getitem__(self, index):
        data_index, subject = self.subjects[index]
        data = self.data[data_index][subject][self.structure][:]

        if self.apply_mask:
            data = self.mask(
                data,
                mask_prob=self.mask_prob,
                vocab_size=self.vocab_size,
                mask_token_id=self.mask_token_id,
                mask_unit=self.mask_unit,
                sep_token_id=self.sep_token_id,
            )

        ret = {
            "id": subject,
            "input_ids": torch.LongTensor(data[:, 0, :]),
            "type_ids": torch.LongTensor(data[:, 1, :]),
            "dpe_ids": torch.LongTensor(data[:, 2, :]),
        }

        if self.label:
            for i, task in enumerate(self.tasks):
                ret[task] = self.labels[self.subjects[index]][task]
                if isinstance(ret[task], str):
                    ret[task] = eval(ret[task])
                # for multi-label classification, where the label is given by a list of class indices
                if isinstance(ret[task], list):
                    ret[task] = list(map(int, ret[task]))
                    num_label = self.num_labels[i]
                    label = np.zeros(num_label, dtype=np.int16)
                    label[ret[task]] = 1
                    ret[task] = torch.tensor(label)
                else:
                    if np.isnan(ret[task]) or ret[task] < 0:
                        ret[task] = self.ignore_index
                    ret[task] = torch.tensor(ret[task])
        return ret


class FlattenedGenHPFDataset(GenHPFDataset):
    def __init__(
        self,
        manifest_paths: List[str],
        label: bool = False,
        tasks: List[str] = None,
        num_labels: List[int] = None,
        **kwargs,
    ):
        kwargs.pop("structure", None)
        super().__init__(manifest_paths=manifest_paths, structure="flattened", **kwargs)

        self.label = label
        self.tasks = tasks
        self.num_labels = num_labels

    def sample_crop_indices(self, size, diff):
        if self.mask:
            start = np.random.randint(0, diff + 1)
            end = size - diff + start
        else:
            start = 0
            end = size - diff
        return start, end

    def pad_to_max_size(self, sample, max_len):
        if len(sample) < max_len:
            sample = np.concatenate([sample, np.zeros(max_len - len(sample), dtype=np.int16)])
        else:
            sample = sample[:max_len]
        return sample

    def collator(self, samples):
        samples = [s for s in samples if s["input_ids"] is not None]
        if len(samples) == 0:
            return {}

        if self.simclr:
            input_ids = sum(
                [
                    [s["input_ids"][: len(s["input_ids"]) // 2], s["input_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
            type_ids = sum(
                [
                    [s["type_ids"][: len(s["input_ids"]) // 2], s["type_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
            dpe_ids = sum(
                [
                    [s["dpe_ids"][: len(s["input_ids"]) // 2], s["dpe_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
            if self.apply_mask:
                input_label = sum(
                    [
                        [
                            s["input_label"][: len(s["input_ids"]) // 2],
                            s["input_label"][len(s["input_ids"]) // 2 :],
                        ]
                        for s in samples
                    ],
                    [],
                )
        else:
            input_ids = [s["input_ids"] for s in samples]
            type_ids = [s["type_ids"] for s in samples]
            dpe_ids = [s["dpe_ids"] for s in samples]
            if self.apply_mask:
                input_label = [s["input_label"] for s in samples]
                type_label = [s["type_label"] for s in samples]
                dpe_label = [s["dpe_label"] for s in samples]

        sizes = [s.size(0) for s in input_ids]
        target_size = max(sizes)

        collated_input_ids = input_ids[0].new_zeros((len(input_ids), target_size)).long()
        collated_type_ids = type_ids[0].new_zeros((len(type_ids), target_size)).long()
        collated_dpe_ids = dpe_ids[0].new_zeros((len(dpe_ids), target_size)).long()
        if self.apply_mask:
            collated_input_label = input_label[0].new_zeros((len(input_label), target_size)).long()
            collated_type_label = type_label[0].new_zeros((len(type_label), target_size)).long()
            collated_dpe_label = dpe_label[0].new_zeros((len(dpe_label), target_size)).long()
        for i, size in enumerate(sizes):
            diff = size - target_size
            if diff == 0:
                collated_input_ids[i] = input_ids[i]
                collated_type_ids[i] = type_ids[i]
                collated_dpe_ids[i] = dpe_ids[i]
                if self.apply_mask:
                    collated_input_label[i] = input_label[i]
                    collated_type_label[i] = type_label[i]
                    collated_dpe_label[i] = dpe_label[i]
            elif diff < 0:
                collated_input_ids[i] = torch.cat(
                    [
                        input_ids[i],
                        input_ids[i].new_zeros(
                            -diff,
                        ),
                    ],
                    dim=0,
                )
                collated_type_ids[i] = torch.cat(
                    [
                        type_ids[i],
                        type_ids[i].new_zeros(
                            -diff,
                        ),
                    ],
                    dim=0,
                )
                collated_dpe_ids[i] = torch.cat(
                    [
                        dpe_ids[i],
                        dpe_ids[i].new_zeros(
                            -diff,
                        ),
                    ],
                    dim=0,
                )
                if self.apply_mask:
                    collated_input_label[i] = torch.cat(
                        [
                            input_label[i],
                            input_label[i].new_zeros(
                                -diff,
                            ),
                        ],
                        dim=0,
                    )
                    collated_type_label[i] = torch.cat(
                        [
                            type_label[i],
                            type_label[i].new_zeros(
                                -diff,
                            ),
                        ],
                        dim=0,
                    )
                    collated_dpe_label[i] = torch.cat(
                        [
                            dpe_label[i],
                            dpe_label[i].new_zeros(
                                -diff,
                            ),
                        ],
                        dim=0,
                    )
            else:
                raise ValueError(f"size mismatch, expected <={target_size}, got {size}")

        out = {"id": [s["id"] for s in samples]}
        out["net_input"] = {
            "input_ids": collated_input_ids,
            "type_ids": collated_type_ids,
            "dpe_ids": collated_dpe_ids,
        }

        if self.apply_mask:
            out["input_label"] = collated_input_label
            out["type_label"] = collated_type_label
            out["dpe_label"] = collated_dpe_label

        if self.label:
            label = {}
            for task in self.tasks:
                label[task] = torch.stack([s[task] for s in samples])
            out["label"] = label

        return out

    def __getitem__(self, index):
        data_index, subject = self.subjects[index]
        data = self.data[data_index][subject][self.structure][:]

        if self.apply_mask:
            data, mlm_labels = self.mask(
                data,
                mask_prob=self.mask_prob,
                vocab_size=self.vocab_size,
                mask_token_id=self.mask_token_id,
                mask_unit=self.mask_unit,
                sep_token_id=self.sep_token_id,
            )

        ret = {
            "id": self.subjects[index],
            "input_ids": torch.LongTensor(data[0, :]),
            "type_ids": torch.LongTensor(data[1, :]),
            "dpe_ids": torch.LongTensor(data[2, :]),
        }
        if self.apply_mask:
            ret["input_label"] = torch.LongTensor(mlm_labels[0, :])
            ret["type_label"] = torch.LongTensor(mlm_labels[1, :])
            ret["dpe_label"] = torch.LongTensor(mlm_labels[2, :])

        if self.label:
            for i, task in enumerate(self.tasks):
                ret[task] = self.labels[self.subjects[index]][task]
                if isinstance(ret[task], str):
                    ret[task] = eval(ret[task])
                # for multi-label classification, where the label is given by a list of class indices
                if isinstance(ret[task], list):
                    ret[task] = list(map(int, ret[task]))
                    num_label = self.num_labels[i]
                    label = np.zeros(num_label, dtype=np.int16)
                    label[ret[task]] = 1
                    ret[task] = torch.tensor(label)
                else:
                    if np.isnan(ret[task]) or ret[task] < 0:
                        ret[task] = self.ignore_index
                    ret[task] = torch.tensor(ret[task])

        return ret



import logging
from dataclasses import dataclass, field
from typing import Optional

import torch
import torch.nn as nn
from omegaconf import II

from genhpf.configs import BaseConfig, ChoiceEnum
from genhpf.modules import PositionalEncoding

logger = logging.getLogger(__name__)

GENHPF_MODEL_ARCH_CHOICES = ChoiceEnum(["hierarchical", "flattened"])
GENHPF_AGGREGATOR_ARCH_CHOICES = ChoiceEnum(["transformer", "performer"])
GENHPF_EMBEDDING_METHOD_CHOICES = ChoiceEnum(["code", "text"])


@dataclass
class GenHPFConfig(BaseConfig):
    structure: GENHPF_MODEL_ARCH_CHOICES = field(
        default="hierarchical",
        metadata={"help": "Architecture choice for GenHPF. Choose from hierarchical or flattened"},
    )
    embedding_method: GENHPF_EMBEDDING_METHOD_CHOICES = field(
        default="text", metadata={"help": "Embedding method choice for GenHPF. Choose from code or text"}
    )

    encoder_max_seq_len: int = field(
        default=128,
        metadata={
            "help": "max sequence length for the event encoder, only used when structure is "
            "hierarchical. this is the max number of tokens in an event."
        },
    )
    agg_max_seq_len: int = field(
        default=256,
        metadata={
            "help": "max sequence length for the event aggregator. In the hierarchical structure, "
            "this is the max number of events in a sample. In the flattened structure, this is the"
            "max sequence length of the flattened input."
        },
    )

    # configs for event encoder in hierarchical structure
    encoder_layers: int = field(default=2, metadata={"help": "num encoder layers in the transformer"})
    encoder_embed_dim: int = field(default=128, metadata={"help": "encoder embedding dimension"})
    encoder_ffn_embed_dim: int = field(default=512, metadata={"help": "encoder embedding dimension for FFN"})
    encoder_attention_heads: int = field(
        default=4, metadata={"help": "num attention heads in the transformer"}
    )

    # configs for event aggregator
    agg_arch: GENHPF_AGGREGATOR_ARCH_CHOICES = field(
        default="transformer",
        metadata={
            "help": "Architecture choice for the event aggregator. Choose from transformer or " "performer"
        },
    )
    agg_layers: int = field(default=2, metadata={"help": "num layers in the transformer"})
    agg_embed_dim: int = field(default=128, metadata={"help": "hidden dimension for the event aggregator"})
    agg_ffn_embed_dim: int = field(
        default=512, metadata={"help": "hidden dimension for the FFN in the event aggregator"}
    )
    agg_attention_heads: int = field(
        default=4, metadata={"help": "num attention heads for the event aggregator"}
    )

    dropout: float = field(default=0.2, metadata={"help": "dropout probability"})

    from_pretrained: Optional[str] = field(
        default=None, metadata={"help": "path to the pretrained model if available"}
    )

    vocab_size: int = II("dataset.vocab_size")
    debug: bool = II("common.debug")


class GenHPF(nn.Module):
    def __init__(self, cfg: GenHPFConfig):
        super().__init__()
        self.cfg = cfg

        if cfg.debug:
            cfg.encoder_layers = 1
            cfg.encoder_embed_dim = 32
            cfg.encoder_ffn_embed_dim = 128
            cfg.encoder_attention_heads = 2
            cfg.agg_layers = 1
            cfg.agg_embed_dim = 32
            cfg.agg_ffn_embed_dim = 128
            cfg.agg_attention_heads = 2

        self.structure = cfg.structure
        assert self.structure in GENHPF_MODEL_ARCH_CHOICES

        self.embedding_method = cfg.embedding_method
        assert self.embedding_method in GENHPF_EMBEDDING_METHOD_CHOICES

        self.word_embeddings = nn.Embedding(cfg.vocab_size, cfg.encoder_embed_dim, padding_idx=0)
        if self.embedding_method == "text":
            # we currently use 7 token types and 16 digit places in this version
            self.token_type_vocab_size = 7
            self.digit_place_vocab_size = 16
            self.token_type_embeddings = nn.Embedding(
                self.token_type_vocab_size, cfg.encoder_embed_dim, padding_idx=0
            )
            self.digit_place_embeddings = nn.Embedding(
                self.digit_place_vocab_size, cfg.encoder_embed_dim, padding_idx=0
            )

        max_length = cfg.encoder_max_seq_len if self.structure == "hierarchical" else cfg.agg_max_seq_len
        self.positional_encodings = PositionalEncoding(cfg.encoder_embed_dim, cfg.dropout, max_length)

        self.event_encoder = None
        if self.structure == "hierarchical" and self.embedding_method == "text":
            self.encoder_layer_norm = nn.LayerNorm(cfg.encoder_embed_dim, eps=1e-12)
            encoder_layer = nn.TransformerEncoderLayer(
                cfg.encoder_embed_dim,
                cfg.encoder_attention_heads,
                cfg.encoder_ffn_embed_dim,
                cfg.dropout,
                batch_first=True,
            )
            self.event_encoder = nn.TransformerEncoder(encoder_layer, self.cfg.encoder_layers)
            self.post_encode_proj = nn.Linear(cfg.encoder_embed_dim, cfg.agg_embed_dim)
            self.event_positional_encodings = PositionalEncoding(
                cfg.agg_embed_dim, cfg.dropout, cfg.agg_max_seq_len
            )

        self.agg_layer_norm = nn.LayerNorm(cfg.agg_embed_dim, eps=1e-12)
        if cfg.agg_arch == "transformer":
            agg_layer = nn.TransformerEncoderLayer(
                cfg.agg_embed_dim,
                cfg.agg_attention_heads,
                cfg.agg_ffn_embed_dim,
                cfg.dropout,
                batch_first=True,
            )
            self.event_aggregator = nn.TransformerEncoder(agg_layer, cfg.agg_layers)
        elif cfg.agg_arch == "performer":
            from performer_pytorch import Performer

            self.event_aggregator = Performer(
                dim=cfg.agg_embed_dim,
                depth=cfg.agg_layers,
                heads=cfg.agg_attention_heads,
                dim_head=64,
                nb_features=64,
                reversible=True,
                generalized_attention=True,
                ff_dropout=cfg.dropout,
                attn_dropout=cfg.dropout,
                shift_tokens=True,
            )
        else:
            raise NotImplementedError(f"Unsupported event aggregator architecture: {cfg.agg_arch}")

    @classmethod
    def build_model(cls, cfg):
        """Build a new model instance."""
        if cfg.from_pretrained:
            return cls.from_pretrained(cfg, cfg.from_pretrained)
        else:
            return cls(cfg)

    @classmethod
    def from_pretrained(cls, cfg, checkpoint_path):
        model = cls(cfg)
        state_dict = torch.load(checkpoint_path, map_location="cpu")["model"]

        missing_keys, unexpected_keys = model.load_state_dict(state_dict, strict=False)
        assert len(unexpected_keys) == 0, f"unexpected keys: {unexpected_keys}"

        finetuning_parameter_names = []
        if hasattr(model, "get_finetuning_parameter_names"):
            finetuning_parameter_names = model.get_finetuning_parameter_names()
        missing_keys = list(set(missing_keys) - set(finetuning_parameter_names))
        assert len(missing_keys) == 0, f"missing keys: {missing_keys}"

        logger.info(f"loaded pre-trained model from {checkpoint_path}")

        return model

    def get_logits(self, sample, net_output):
        """Get logits from the model output."""
        raise NotImplementedError

    def get_targets(self, sample, net_output):
        """Get targets from the sample or model output."""
        raise NotImplementedError

    def forward(
        self,
        input_ids: torch.Tensor,
        type_ids: torch.Tensor = None,
        dpe_ids: torch.Tensor = None,
        padding_mask: torch.Tensor = None,
        encoder_only: bool = False,
        **kwargs,
    ):
        x = self.word_embeddings(input_ids)
        if self.embedding_method == "text":
            if type_ids is not None:
                x += self.token_type_embeddings(type_ids)
            if dpe_ids is not None:
                x += self.digit_place_embeddings(dpe_ids)

        if self.structure == "hierarchical":
            assert input_ids.ndim == 3  # (batch, num_events, num_words)
            batch_size, num_events = input_ids.shape[0], input_ids.shape[1]

            x = x.view(batch_size * num_events, -1, self.cfg.encoder_embed_dim)
            x = self.positional_encodings(x)
            x = self.encoder_layer_norm(x)
            if padding_mask is None:
                padding_mask = input_ids.view(batch_size * num_events, -1).eq(0).to(x.device)
            # x: (batch * num_events, num_words, embed_dim)
            x = self.event_encoder(x, src_key_padding_mask=padding_mask)

            if padding_mask.any():
                x[padding_mask] = 0
            x = torch.div(x.sum(dim=1), (x != 0).sum(dim=1))
            x = self.post_encode_proj(x).view(batch_size, num_events, -1)  # (batch, num_events, embed_dim)

            padding_mask = input_ids[:, :, 1].eq(0).to(x.device)

            x = self.event_positional_encodings(x)
        else:
            assert input_ids.ndim == 2  # (batch, seq_len)
            x = self.positional_encodings(x)  # (batch, seq_len, embed_dim)
            if padding_mask is None:
                padding_mask = input_ids.eq(0).to(x.device)

        if encoder_only:
            return {
                "x": x,
                "padding_mask": padding_mask,
            }

        # breakpoint()
        x = self.event_aggregator(x, src_key_padding_mask=padding_mask)

        return x, padding_mask



import glob
import logging
import os
from datetime import datetime

import pandas as pd
import pyspark.sql.functions as F
from ehrs import EHR, register_ehr

logger = logging.getLogger(__name__)


@register_ehr("mimiciii")
class MIMICIII(EHR):
    def __init__(self, cfg):
        super().__init__(cfg)

        self.ehr_name = "mimiciii"

        if self.data_dir is None:
            self.data_dir = os.path.join(self.cache_dir, self.ehr_name)

            if not os.path.exists(self.data_dir):
                logger.info(
                    "Data is not found so try to download from the internet. "
                    "Note that this is a restricted-access resource. "
                    "Please log in to physionet.org with a credentialed user."
                )
                self.download_ehr_from_url(
                    url="https://physionet.org/files/mimiciii/1.4/", dest=self.data_dir
                )

        logger.info("Data directory is set to {}".format(self.data_dir))

        if self.ccs_path is None:
            self.ccs_path = os.path.join(self.cache_dir, "ccs_multi_dx_tool_2015.csv")

            if not os.path.exists(self.ccs_path):
                logger.info("`ccs_multi_dx_tool_2015.csv` is not found so try to download from the internet.")
                self.download_ccs_from_url(self.cache_dir)

        if self.ext is None:
            self.ext = self.infer_data_extension()

        # constants
        self._icustay_fname = "ICUSTAYS" + self.ext
        self._patient_fname = "PATIENTS" + self.ext
        self._admission_fname = "ADMISSIONS" + self.ext
        self._diagnosis_fname = "DIAGNOSES_ICD" + self.ext

        # XXX more features? user choice?
        self.tables = [
            {
                "fname": "LABEVENTS" + self.ext,
                "timestamp": "CHARTTIME",
                "timeoffsetunit": "abs",
                "exclude": ["ROW_ID", "SUBJECT_ID"],
                "code": ["ITEMID"],
                "desc": ["D_LABITEMS" + self.ext],
                "desc_key": ["LABEL"],
            },
            {
                "fname": "PRESCRIPTIONS" + self.ext,
                "timestamp": "STARTDATE",
                "timeoffsetunit": "abs",
                "exclude": ["ENDDATE", "GSN", "NDC", "ROW_ID", "SUBJECT_ID"],
            },
            {
                "fname": "INPUTEVENTS_MV" + self.ext,
                "timestamp": "STARTTIME",
                "timeoffsetunit": "abs",
                "exclude": [
                    "ENDTIME",
                    "STORETIME",
                    "CGID",
                    "ORDERID",
                    "LINKORDERID",
                    "ROW_ID",
                    "SUBJECT_ID",
                    "CONTINUEINNEXTDEPT",
                    "CANCELREASON",
                    "STATUSDESCRIPTION",
                    "COMMENTS_CANCELEDBY",
                    "COMMENTS_DATE",
                ],
                "code": ["ITEMID"],
                "desc": ["D_ITEMS" + self.ext],
                "desc_key": ["LABEL"],
            },
            {
                "fname": "INPUTEVENTS_CV" + self.ext,
                "timestamp": "CHARTTIME",
                "timeoffsetunit": "abs",
                "exclude": [
                    "STORETIME",
                    "CGID",
                    "ORDERID",
                    "LINKORDERID",
                    "ROW_ID",
                    "STOPPED",
                    "SUBJECT_ID",
                ],
                "code": ["ITEMID"],
                "desc": ["D_ITEMS" + self.ext],
                "desc_key": ["LABEL"],
            },
        ]

        if self.feature == "select":
            extra_exclude_feature_dict = {
                "LABEVENTS" + self.ext: ["VALUE", "FLAG"],
                "PRESCRIPTIONS"
                + self.ext: [
                    "DRUG_TYPE",
                    "DRUG_NAME_POE",
                    "DRUG_NAME_GENERIC",
                    "FORMULARY_DRUG_CD",
                    "FORM_VAL_DISP",
                    "FORM_UNIT_DISP",
                ],
                "INPUTEVENTS_MV"
                + self.ext: [
                    "AMOUNT",
                    "AMOUNTUOM",
                    "ORDERCATEGORYNAME",
                    "SECONDARYORDERCATEGORYNAME",
                    "ORDERCOMPONENTTYPEDESCRIPTION",
                    "ORDERCATEGORYDESCRIPTION",
                    "PATIENTWEIGHT",
                    "TOTALAMOUNT",
                    "TOTALAMOUNTUOM",
                    "ISOPENBAG",
                    "COMMENTS_EDITEDBY",
                    "ORIGINALAMOUNT",
                    "ORIGINALRATE",
                ],
                "INPUTEVENTS_CV"
                + self.ext: [
                    "AMOUNT",
                    "AMOUNTUOM",
                    "NEWBOTTLE",
                    "ORIGINALAMOUNT",
                    "ORIGINALRATE",
                    "ORIGIANLAMOUNTUOM",
                    "ORIGINALRATE",
                    "ORIGINALRATEUOM",
                    "ORIGINALSITE",
                ],
            }

            for table in self.tables:
                if table["fname"] in extra_exclude_feature_dict.keys():
                    exclude_target_list = extra_exclude_feature_dict[table["fname"]]
                    table["exclude"].extend(exclude_target_list)

        if self.emb_type == "codebase":
            feature_types_for_codebase_emb_dict = {
                "LABEVENTS"
                + self.ext: {
                    "numeric_feat": ["VALUE", "VALUENUM"],
                    "categorical_feat": [],
                    "code_feat": ["ITEMID"],
                },
                "PRESCRIPTIONS"
                + self.ext: {
                    "numeric_feat": ["DOSE_VAL_RX", "FORM_VAL_DISP"],
                    "categorical_feat": [],
                    "code_feat": ["DRUG"],
                },
                "INPUTEVENTS_MV"
                + self.ext: {
                    "numeric_feat": [
                        "AMOUNT",
                        "RATE",
                        "PATIENTWEIGHT",
                        "TOTALAMOUNT",
                        "ORIGINALAMOUNT",
                        "ORIGINALRATE",
                    ],
                    "categorical_feat": ["ISOPENBAG", "CONTINUEINNEXTDEPT", "CANCELREASON", "VALUECOUNTS"],
                    "code_feat": ["ITEMID"],
                },
                "INPUTEVENTS_CV"
                + self.ext: {
                    "numeric_feat": [
                        "AMOUNT",
                        "RATE",
                        "PATIENTWEIGHT",
                        "TOTALAMOUNT",
                        "ORIGINALAMOUNT",
                        "ORIGINALRATE",
                    ],
                    "categorical_feat": ["ISOPENBAG", "CONTINUEINNEXTDEPT", "CANCELREASON", "VALUECOUNTS"],
                    "code_feat": ["ITEMID"],
                },
            }

            for table in self.tables:
                if table["fname"] in feature_types_for_codebase_emb_dict.keys():
                    feature_dict = feature_types_for_codebase_emb_dict[table["fname"]]
                    table.update(feature_dict)

        if self.creatinine or self.bilirubin or self.platelets or self.wbc:
            self.task_itemids = {
                "creatinine": {
                    "fname": "LABEVENTS" + self.ext,
                    "timestamp": "CHARTTIME",
                    "timeoffsetunit": "abs",
                    "exclude": ["ROW_ID", "SUBJECT_ID", "VALUE", "VALUEUOM", "FLAG"],
                    "code": ["ITEMID"],
                    "value": ["VALUENUM"],
                    "itemid": [50912],
                },
                "bilirubin": {
                    "fname": "LABEVENTS" + self.ext,
                    "timestamp": "CHARTTIME",
                    "timeoffsetunit": "abs",
                    "exclude": ["ROW_ID", "SUBJECT_ID", "VALUE", "VALUEUOM", "FLAG"],
                    "code": ["ITEMID"],
                    "value": ["VALUENUM"],
                    "itemid": [50885],
                },
                "platelets": {
                    "fname": "LABEVENTS" + self.ext,
                    "timestamp": "CHARTTIME",
                    "timeoffsetunit": "abs",
                    "exclude": ["ROW_ID", "SUBJECT_ID", "VALUE", "VALUEUOM", "FLAG"],
                    "code": ["ITEMID"],
                    "value": ["VALUENUM"],
                    "itemid": [51265],
                },
                "wbc": {
                    "fname": "LABEVENTS" + self.ext,
                    "timestamp": "CHARTTIME",
                    "timeoffsetunit": "abs",
                    "exclude": ["ROW_ID", "SUBJECT_ID", "VALUE", "FLAG"],
                    "code": ["ITEMID"],
                    "value": ["VALUENUM"],
                    "itemid": [51300, 51301],
                },
                "dialysis": {
                    "tables": {
                        "chartevents": {
                            "fname": "CHARTEVENTS" + self.ext,
                            "timestamp": "CHARTTIME",
                            "timeoffsetunit": "abs",
                            "include": [
                                "ICUSTAY_ID",
                                "SUBJECT_ID",
                                "ITEMID",
                                "VALUE",
                                "VALUENUM",
                                "CHARTTIME",
                                "ERROR",
                            ],
                            "itemid": {
                                "cv_ce": [
                                    152,
                                    148,
                                    149,
                                    146,
                                    147,
                                    151,
                                    150,
                                    7949,
                                    229,
                                    235,
                                    241,
                                    247,
                                    253,
                                    259,
                                    265,
                                    271,
                                    582,
                                    466,
                                    917,
                                    927,
                                    6250,
                                ],
                                "mv_ce": [
                                    226118,
                                    227357,
                                    225725,
                                    226499,
                                    224154,
                                    225810,
                                    227639,
                                    225183,
                                    227438,
                                    224191,
                                    225806,
                                    225807,
                                    228004,
                                    228005,
                                    228006,
                                    224144,
                                    224145,
                                    224149,
                                    224150,
                                    224151,
                                    224152,
                                    224153,
                                    224404,
                                    224406,
                                    226457,
                                    225959,
                                    224135,
                                    224139,
                                    224146,
                                    225323,
                                    225740,
                                    225776,
                                    225951,
                                    225952,
                                    225953,
                                    225954,
                                    225956,
                                    225958,
                                    225961,
                                    225963,
                                    225965,
                                    225976,
                                    225977,
                                    227124,
                                    227290,
                                    227638,
                                    227640,
                                    227753,
                                ],
                            },
                        },
                        "inputevents_cv": {
                            "fname": "INPUTEVENTS_CV" + self.ext,
                            "timestamp": "CHARTTIME",
                            "timeoffsetunit": "abs",
                            "include": ["SUBJECT_ID", "ITEMID", "AMOUNT", "CHARTTIME"],
                            "itemid": {
                                "cv_ie": [
                                    40788,
                                    40907,
                                    41063,
                                    41147,
                                    41307,
                                    41460,
                                    41620,
                                    41711,
                                    41791,
                                    41792,
                                    42562,
                                    43829,
                                    44037,
                                    44188,
                                    44526,
                                    44527,
                                    44584,
                                    44591,
                                    44698,
                                    44927,
                                    44954,
                                    45157,
                                    45268,
                                    45352,
                                    45353,
                                    46012,
                                    46013,
                                    46172,
                                    46173,
                                    46250,
                                    46262,
                                    46292,
                                    46293,
                                    46311,
                                    46389,
                                    46574,
                                    46681,
                                    46720,
                                    46769,
                                    46773,
                                ]
                            },
                        },
                        "outputevents": {
                            "fname": "OUTPUTEVENTS" + self.ext,
                            "timestamp": "CHARTTIME",
                            "timeoffsetunit": "abs",
                            "include": ["SUBJECT_ID", "ITEMID", "VALUE", "CHARTTIME"],
                            "itemid": {
                                "cv_oe": [
                                    40386,
                                    40425,
                                    40426,
                                    40507,
                                    40613,
                                    40624,
                                    40690,
                                    40745,
                                    40789,
                                    40881,
                                    40910,
                                    41016,
                                    41034,
                                    41069,
                                    41112,
                                    41250,
                                    41374,
                                    41417,
                                    41500,
                                    41527,
                                    41623,
                                    41635,
                                    41713,
                                    41750,
                                    41829,
                                    41842,
                                    41897,
                                    42289,
                                    42388,
                                    42464,
                                    42524,
                                    42536,
                                    42868,
                                    42928,
                                    42972,
                                    43016,
                                    43052,
                                    43098,
                                    43115,
                                    43687,
                                    43941,
                                    44027,
                                    44085,
                                    44193,
                                    44199,
                                    44216,
                                    44286,
                                    44567,
                                    44843,
                                    44845,
                                    44857,
                                    44901,
                                    44943,
                                    45479,
                                    45828,
                                    46230,
                                    46232,
                                    46394,
                                    46464,
                                    46712,
                                    46713,
                                    46715,
                                    46741,
                                ],
                            },
                        },
                        "inputevents_mv": {
                            "fname": "INPUTEVENTS_MV" + self.ext,
                            "timestamp": "STARTTIME",
                            "timeoffsetunit": "abs",
                            "include": ["SUBJECT_ID", "ITEMID", "AMOUNT", "STARTTIME"],
                            "itemid": {"mv_ie": [227536, 227525]},
                        },
                        "datetimeevents": {
                            "fname": "DATETIMEEVENTS" + self.ext,
                            "timestamp": "CHARTTIME",
                            "timeoffsetunit": "abs",
                            "include": ["SUBJECT_ID", "ITEMID", "CHARTTIME"],
                            "itemid": {"mv_de": [225318, 225319, 225321, 225322, 225324]},
                        },
                        "procedureevents_mv": {
                            "fname": "PROCEDUREEVENTS_MV" + self.ext,
                            "timestamp": "STARTTIME",
                            "timeoffsetunit": "abs",
                            "include": ["SUBJECT_ID", "ITEMID", "STARTTIME"],
                            "itemid": {
                                "mv_pe": [225441, 225802, 225803, 225805, 224270, 225809, 225955, 225436]
                            },
                        },
                    }
                },
            }

        self.disch_map_dict = {
            "DISC-TRAN CANCER/CHLDRN H": "Other",
            "DISC-TRAN TO FEDERAL HC": "Other",
            "DISCH-TRAN TO PSYCH HOSP": "Other",
            "HOME": "Home",
            "HOME HEALTH CARE": "Home",
            "HOME WITH HOME IV PROVIDR": "Home",
            "HOSPICE-HOME": "Other",
            "HOSPICE-MEDICAL FACILITY": "Other",
            "ICF": "Other",
            "IN_ICU_MORTALITY": "IN_ICU_MORTALITY",
            "LEFT AGAINST MEDICAL ADVI": "Other",
            "LONG TERM CARE HOSPITAL": "Other",
            "OTHER FACILITY": "Other",
            "REHAB/DISTINCT PART HOSP": "Rehabilitation",
            "SHORT TERM HOSPITAL": "Other",
            "SNF": "Skilled Nursing Facility",
            "SNF-MEDICAID ONLY CERTIF": "Skilled Nursing Facility",
            "Death": "Death",
        }
        self._icustay_key = "ICUSTAY_ID"
        self._hadm_key = "HADM_ID"
        self._patient_key = "SUBJECT_ID"

        self._determine_first_icu = "INTIME"

    def build_cohorts(self, cached=False):
        icustays = pd.read_csv(os.path.join(self.data_dir, self.icustay_fname))

        icustays = self.make_compatible(icustays)
        self.icustays = icustays

        cohorts = super().build_cohorts(icustays, cached=cached)

        return cohorts

    def prepare_tasks(self, cohorts, spark, cached=False):
        if cohorts is None and cached:
            labeled_cohorts = self.load_from_cache(self.ehr_name + ".cohorts.labeled")
            if labeled_cohorts is not None:
                return labeled_cohorts

        labeled_cohorts = super().prepare_tasks(cohorts, spark, cached)

        if self.diagnosis:
            logger.info("Start labeling cohorts for diagnosis prediction.")
            # define diagnosis prediction task
            diagnoses = pd.read_csv(os.path.join(self.data_dir, self.diagnosis_fname))

            ccs_dx = pd.read_csv(self.ccs_path)
            ccs_dx["'ICD-9-CM CODE'"] = ccs_dx["'ICD-9-CM CODE'"].str[1:-1].str.strip()
            ccs_dx["'CCS LVL 1'"] = ccs_dx["'CCS LVL 1'"].str[1:-1]
            lvl1 = {x: int(y) - 1 for _, (x, y) in ccs_dx[["'ICD-9-CM CODE'", "'CCS LVL 1'"]].iterrows()}
            diagnoses["diagnosis"] = diagnoses["ICD9_CODE"].map(lvl1)

            diagnoses = diagnoses[(diagnoses["diagnosis"].notnull()) & (diagnoses["diagnosis"] != 14)]
            diagnoses.loc[diagnoses["diagnosis"] >= 14, "diagnosis"] -= 1
            diagnoses = diagnoses.groupby(self.hadm_key)["diagnosis"].agg(lambda x: list(set(x))).to_frame()
            labeled_cohorts = labeled_cohorts.merge(diagnoses, on=self.hadm_key, how="inner")

            # labeled_cohorts = labeled_cohorts.drop(columns=["ICD9_CODE"])

            logger.info(
                "Done preparing diagnosis prediction for the given cohorts, Cohort Numbers: {}".format(
                    len(labeled_cohorts)
                )
            )

            self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled")

        if self.bilirubin or self.platelets or self.creatinine or self.wbc:
            logger.info("Start labeling cohorts for clinical task prediction.")

            labeled_cohorts = spark.createDataFrame(labeled_cohorts)

            if self.bilirubin:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "bilirubin", spark)

            if self.platelets:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "platelets", spark)

            if self.creatinine:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "creatinine", spark)

            if self.wbc:
                labeled_cohorts = self.clinical_task(labeled_cohorts, "wbc", spark)
            # labeled_cohorts = labeled_cohorts.toPandas()

            # self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled.clinical_tasks")

            logger.info("Done preparing clinical task prediction for the given cohorts")

        # self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled")
        return labeled_cohorts

    def make_compatible(self, icustays):
        patients = pd.read_csv(os.path.join(self.data_dir, self.patient_fname))
        admissions = pd.read_csv(os.path.join(self.data_dir, self.admission_fname))

        # prepare icustays according to the appropriate format
        icustays = icustays[icustays["FIRST_CAREUNIT"] == icustays["LAST_CAREUNIT"]]

        icustays["INTIME"] = pd.to_datetime(icustays["INTIME"], infer_datetime_format=True, utc=True)

        icustays["OUTTIME"] = pd.to_datetime(icustays["OUTTIME"], infer_datetime_format=True, utc=True)
        icustays = icustays.drop(columns=["ROW_ID"])

        # merge icustays with patients to get DOB
        patients["DOB"] = pd.to_datetime(patients["DOB"], infer_datetime_format=True, utc=True)
        patients = patients[patients["SUBJECT_ID"].isin(icustays["SUBJECT_ID"])]
        patients = patients.drop(columns=["ROW_ID"])[["DOB", "SUBJECT_ID"]]
        icustays = icustays.merge(patients, on="SUBJECT_ID", how="left")

        def calculate_age(birth: datetime, now: datetime):
            age = now.year - birth.year
            if now.month < birth.month:
                age -= 1
            elif (now.month == birth.month) and (now.day < birth.day):
                age -= 1

            return age

        icustays["AGE"] = icustays.apply(lambda x: calculate_age(x["DOB"], x["INTIME"]), axis=1)

        # merge with admissions to get discharge information
        icustays = pd.merge(
            icustays.reset_index(drop=True),
            admissions[["HADM_ID", "DISCHARGE_LOCATION", "DEATHTIME", "DISCHTIME"]],
            how="left",
            on="HADM_ID",
        )
        icustays["DISCHARGE_LOCATION"].replace("DEAD/EXPIRED", "Death", inplace=True)

        icustays["DISCHTIME"] = pd.to_datetime(icustays["DISCHTIME"], infer_datetime_format=True, utc=True)

        icustays["IN_ICU_MORTALITY"] = (
            (icustays["INTIME"] < icustays["DISCHTIME"])
            & (icustays["DISCHTIME"] <= icustays["OUTTIME"])
            & (icustays["DISCHARGE_LOCATION"] == "Death")
        )
        icustays["DISCHARGE_LOCATION"] = icustays["DISCHARGE_LOCATION"].map(self.disch_map_dict)

        icustays.rename(columns={"DISCHARGE_LOCATION": "HOS_DISCHARGE_LOCATION"}, inplace=True)
        icustays["DISCHTIME"] = (icustays["DISCHTIME"] - icustays["INTIME"]).dt.total_seconds() // 60
        icustays["OUTTIME"] = (icustays["OUTTIME"] - icustays["INTIME"]).dt.total_seconds() // 60
        return icustays

    def clinical_task(self, cohorts, task, spark):
        fname = self.task_itemids[task]["fname"]
        timestamp = self.task_itemids[task]["timestamp"]
        timeoffsetunit = self.task_itemids[task]["timeoffsetunit"]
        excludes = self.task_itemids[task]["exclude"]
        code = self.task_itemids[task]["code"][0]
        value = self.task_itemids[task]["value"][0]
        itemid = self.task_itemids[task]["itemid"]

        table = spark.read.csv(os.path.join(self.data_dir, fname), header=True)
        table = table.drop(*excludes)
        table = table.filter(F.col(code).isin(itemid)).filter(F.col(value).isNotNull())

        merge = cohorts.join(table, on=self.hadm_key, how="inner")

        if timeoffsetunit == "abs":
            merge = merge.withColumn(timestamp, F.to_timestamp(timestamp))

        # For Creatinine task, eliminate icus if patient went through dialysis treatment
        # before (obs_size + pred_size) timestamp
        # see https://github.com/MIT-LCP/mimic-code/blob/main/mimic-iii/concepts/rrt.sql
        if task == "creatinine":
            dialysis_tables = self.task_itemids["dialysis"]["tables"]
            chartevents = spark.read.csv(os.path.join(self.data_dir, "CHARTEVENTS" + self.ext), header=True)
            inputevents_cv = spark.read.csv(
                os.path.join(self.data_dir, "INPUTEVENTS_CV" + self.ext), header=True
            )
            outputevents = spark.read.csv(os.path.join(self.data_dir, "OUTPUTEVENTS" + self.ext), header=True)
            inputevents_mv = spark.read.csv(
                os.path.join(self.data_dir, "INPUTEVENTS_MV" + self.ext), header=True
            )
            datetimeevents = spark.read.csv(
                os.path.join(self.data_dir, "DATETIMEEVENTS" + self.ext), header=True
            )
            procedureevents_mv = spark.read.csv(
                os.path.join(self.data_dir, "PROCEDUREEVENTS_MV" + self.ext), header=True
            )
            icustays = spark.read.csv(os.path.join(self.data_dir, "ICUSTAYS" + self.ext), header=True)

            chartevents = chartevents.select(*dialysis_tables["chartevents"]["include"])
            inputevents_cv = inputevents_cv.select(*dialysis_tables["inputevents_cv"]["include"])
            outputevents = outputevents.select(*dialysis_tables["outputevents"]["include"])
            inputevents_mv = inputevents_mv.select(*dialysis_tables["inputevents_mv"]["include"])
            datetimeevents = datetimeevents.select(*dialysis_tables["datetimeevents"]["include"])
            procedureevents_mv = procedureevents_mv.select(*dialysis_tables["procedureevents_mv"]["include"])

            # Filter dialysis related tables with dialysis condition #TODO: check dialysis condition
            cv_ce = (
                chartevents.filter(F.col("ITEMID").isin(dialysis_tables["chartevents"]["itemid"]["cv_ce"]))
                .filter(F.col("VALUE").isNotNull())
                .filter((F.col("ERROR").isNull()) | (F.col("ERROR") == 0))
                .filter(
                    (
                        (F.col("ITEMID").isin([152, 148, 149, 146, 147, 151, 150]))
                        & (F.col("VALUE").isNotNull())
                    )
                    | (
                        (F.col("ITEMID").isin([229, 235, 241, 247, 253, 259, 265, 271]))
                        & (F.col("VALUE") == "Dialysis Line")
                    )
                    | ((F.col("ITEMID") == 466) & (F.col("VALUE") == "Dialysis RN"))
                    | ((F.col("ITEMID") == 927) & (F.col("VALUE") == "Dialysis Solutions"))
                    | ((F.col("ITEMID") == 6250) & (F.col("VALUE") == "dialys"))
                    | (
                        (F.col("ITEMID") == 917)
                        & (
                            F.col("VALUE").isin(
                                [
                                    "+ INITIATE DIALYSIS",
                                    "BLEEDING FROM DIALYSIS CATHETER",
                                    "FAILED DIALYSIS CATH.",
                                    "FEBRILE SYNDROME;DIALYSIS",
                                    "HYPOTENSION WITH HEMODIALYSIS",
                                    "HYPOTENSION.GLOGGED DIALYSIS",
                                    "INFECTED DIALYSIS CATHETER",
                                ]
                            )
                        )
                    )
                    | (
                        (F.col("ITEMID") == 582)
                        & (
                            F.col("VALUE").isin(
                                [
                                    "CAVH Start",
                                    "CAVH D/C",
                                    "CVVHD Start",
                                    "CVVHD D/C",
                                    "Hemodialysis st",
                                    "Hemodialysis end",
                                ]
                            )
                        )
                    )
                )
            )
            icustays = icustays.filter(F.col("DBSOURCE") == "carevue").select(self.icustay_key)
            cv_ce = cv_ce.join(icustays, on=self.icustay_key, how="inner")
            cv_ie = inputevents_cv.filter(
                F.col("ITEMID").isin(dialysis_tables["inputevents_cv"]["itemid"]["cv_ie"])
            ).filter(F.col("AMOUNT") > 0)
            cv_oe = outputevents.filter(
                F.col("ITEMID").isin(dialysis_tables["outputevents"]["itemid"]["cv_oe"])
            ).filter(F.col("VALUE") > 0)
            mv_ce = (
                chartevents.filter(F.col("ITEMID").isin(dialysis_tables["chartevents"]["itemid"]["mv_ce"]))
                .filter(F.col("VALUENUM") > 0)
                .filter((F.col("ERROR").isNull()) | (F.col("ERROR") == 0))
            )
            mv_ie = inputevents_mv.filter(
                F.col("ITEMID").isin(dialysis_tables["inputevents_mv"]["itemid"]["mv_ie"])
            ).filter(F.col("AMOUNT") > 0)
            mv_de = datetimeevents.filter(
                F.col("ITEMID").isin(dialysis_tables["datetimeevents"]["itemid"]["mv_de"])
            )
            mv_pe = procedureevents_mv.filter(
                F.col("ITEMID").isin(dialysis_tables["procedureevents_mv"]["itemid"]["mv_pe"])
            )

            def dialysis_time(table, timecolumn):
                return table.withColumn("_DIALYSIS_TIME", F.to_timestamp(timecolumn)).select(
                    self.patient_key, "_DIALYSIS_TIME"
                )

            cv_ce, cv_ie, cv_oe, mv_ce, mv_ie, mv_de, mv_pe = (
                dialysis_time(cv_ce, "CHARTTIME"),
                dialysis_time(cv_ie, "CHARTTIME"),
                dialysis_time(cv_oe, "CHARTTIME"),
                dialysis_time(mv_ce, "CHARTTIME"),
                dialysis_time(mv_ie, "STARTTIME"),
                dialysis_time(mv_de, "CHARTTIME"),
                dialysis_time(mv_pe, "STARTTIME"),
            )

            dialysis = cv_ce.union(cv_ie).union(cv_oe).union(mv_ce).union(mv_ie).union(mv_de).union(mv_pe)
            dialysis = dialysis.groupBy(self.patient_key).agg(F.min("_DIALYSIS_TIME").alias("_DIALYSIS_TIME"))
            merge = merge.join(dialysis, on=self.patient_key, how="left")
            merge = merge.filter(F.isnull("_DIALYSIS_TIME") | (F.col("_DIALYSIS_TIME") > F.col(timestamp)))
            merge = merge.drop("_DIALYSIS_TIME")

        if timeoffsetunit == "abs":
            merge = merge.withColumn(
                timestamp, F.round((F.col(timestamp).cast("long") - F.col("INTIME").cast("long")) / 60)
            )

        # Cohort with events within (obs_size + gap_size) - (obs_size + pred_size)
        merge = merge.filter(((self.obs_size + self.gap_size) * 60) <= F.col(timestamp)).filter(
            ((self.obs_size + self.pred_size) * 60) >= F.col(timestamp)
        )

        # Average value of events
        value_agg = merge.groupBy(self.icustay_key).agg(
            F.mean(value).alias("avg_value")
        )  # TODO: mean/min/max?

        # Labeling
        if task == "bilirubin":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 1.2, 0)
                .when((value_agg.avg_value >= 1.2) & (value_agg.avg_value < 2.0), 1)
                .when((value_agg.avg_value >= 2.0) & (value_agg.avg_value < 6.0), 2)
                .when((value_agg.avg_value >= 6.0) & (value_agg.avg_value < 12.0), 3)
                .when(value_agg.avg_value >= 12.0, 4),
            )
        elif task == "platelets":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value >= 150, 0)
                .when((value_agg.avg_value >= 100) & (value_agg.avg_value < 150), 1)
                .when((value_agg.avg_value >= 50) & (value_agg.avg_value < 100), 2)
                .when((value_agg.avg_value >= 20) & (value_agg.avg_value < 50), 3)
                .when(value_agg.avg_value < 20, 4),
            )

        elif task == "creatinine":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 1.2, 0)
                .when((value_agg.avg_value >= 1.2) & (value_agg.avg_value < 2.0), 1)
                .when((value_agg.avg_value >= 2.0) & (value_agg.avg_value < 3.5), 2)
                .when((value_agg.avg_value >= 3.5) & (value_agg.avg_value < 5), 3)
                .when(value_agg.avg_value >= 5, 4),
            )

        elif task == "wbc":
            value_agg = value_agg.withColumn(
                task,
                F.when(value_agg.avg_value < 4, 0)
                .when((value_agg.avg_value >= 4) & (value_agg.avg_value <= 12), 1)
                .when((value_agg.avg_value > 12), 2),
            )

        cohorts = cohorts.join(value_agg.select(self.icustay_key, task), on=self.icustay_key, how="left")

        return cohorts

    def infer_data_extension(self) -> str:
        if len(glob.glob(os.path.join(self.data_dir, "*.csv.gz"))) == 26:
            ext = ".csv.gz"
        elif len(glob.glob(os.path.join(self.data_dir, "*.csv"))) == 26:
            ext = ".csv"
        else:
            raise AssertionError(
                "Provided data directory is not correct. Please check if --data is correct. "
                "--data: {}".format(self.data_dir)
            )

        logger.info("Data extension is set to '{}'".format(ext))

        return ext



from typing import List

from genhpf.configs import Config

from .dataset import BaseDataset
from .genhpf_dataset import FlattenedGenHPFDataset, HierarchicalGenHPFDataset
from .meds_dataset import HierarchicalMEDSDataset

__all__ = ["BaseDataset", "HierarchicalGenHPFDataset", "FlattenedGenHPFDataset", "HierarchicalMEDSDataset"]


def load_dataset(
    data_path: str,
    subsets: List[str],
    cfg: Config,
):
    dataset_cfg = cfg.dataset
    model_cfg = cfg.model
    criterion_cfg = cfg.criterion

    manifest_paths = [f"{data_path}/{subset.strip()}.tsv" for subset in subsets]

    if dataset_cfg.data_format == "genhpf":
        if model_cfg.structure == "hierarchical":
            dataset = HierarchicalGenHPFDataset(
                manifest_paths=manifest_paths,
                label=dataset_cfg.label,
                tasks=getattr(criterion_cfg, "task_names", None),
                num_labels=getattr(criterion_cfg, "num_labels", None),
                vocab_size=dataset_cfg.vocab_size,
                pad_token_id=dataset_cfg.pad_token_id,
                sep_token_id=dataset_cfg.sep_token_id,
                ignore_index=dataset_cfg.ignore_index,
                apply_mask=dataset_cfg.apply_mask or "mlm" in model_cfg._name,
                mask_token_id=dataset_cfg.mask_token_id,
                mask_prob=dataset_cfg.mask_prob,
                mask_unit=dataset_cfg.mask_unit,
                simclr="simclr" in model_cfg._name,
                dummy_token_id=dataset_cfg.dummy_token_id,
            )
        else:
            dataset = FlattenedGenHPFDataset(
                manifest_paths=manifest_paths,
                label=dataset_cfg.label,
                tasks=getattr(criterion_cfg, "task_names", None),
                num_labels=getattr(criterion_cfg, "num_labels", None),
                vocab_size=dataset_cfg.vocab_size,
                pad_token_id=dataset_cfg.pad_token_id,
                sep_token_id=dataset_cfg.sep_token_id,
                ignore_index=dataset_cfg.ignore_index,
                apply_mask=dataset_cfg.apply_mask or "mlm" in model_cfg._name,
                mask_token_id=dataset_cfg.mask_token_id,
                mask_prob=dataset_cfg.mask_prob,
                mask_unit=dataset_cfg.mask_unit,
                simclr="simclr" in model_cfg._name,
            )
    elif dataset_cfg.data_format == "meds":
        assert model_cfg.structure == "hierarchical", (
            "we currently only support hierarchical structure for MEDS dataset."
            " please set model.structure to 'hierarchical'"
        )
        dataset = HierarchicalMEDSDataset(
            manifest_paths=manifest_paths,
            max_events=model_cfg.agg_max_seq_len,
            label=dataset_cfg.label,
            tasks=getattr(criterion_cfg, "task_names", None),
            num_labels=getattr(criterion_cfg, "num_labels", None),
            structure=model_cfg.structure,
            vocab_size=dataset_cfg.vocab_size,
            pad_token_id=dataset_cfg.pad_token_id,
            sep_token_id=dataset_cfg.sep_token_id,
            ignore_index=dataset_cfg.ignore_index,
            apply_mask=dataset_cfg.apply_mask or "mlm" in model_cfg._name,
            mask_token_id=dataset_cfg.mask_token_id,
            mask_prob=dataset_cfg.mask_prob,
            mask_unit=dataset_cfg.mask_unit,
            simclr="simclr" in model_cfg._name,
            dummy_token_id=dataset_cfg.dummy_token_id,
            debug=cfg.common.debug,
        )
    else:
        raise NotImplementedError(f"unsupported data format: {dataset_cfg.data_format}")

    return dataset



"""isort:skip_file"""

from importlib.metadata import PackageNotFoundError, version

__package_name__ = "genhpf"
try:
    __version__ = version(__package_name__)
except PackageNotFoundError:
    __version__ = "unknown"



import logging
import os
from typing import List, Union

import h5pickle
import numpy as np
import torch

from genhpf.datasets.dataset import BaseDataset

logger = logging.getLogger(__name__)


class MEDSDataset(BaseDataset):
    def __init__(
        self,
        manifest_paths: List[str],
        structure: str = "hierarchical",
        vocab_size: int = 28996,
        pad_token_id: int = 0,
        sep_token_id: int = 102,
        ignore_index: int = -100,
        apply_mask: bool = False,
        mask_token_id: int = 103,
        mask_prob: float = 0,
        mask_unit: str = "individual",
        simclr: bool = False,
        debug: bool = False,
        **kwargs,
    ):
        super().__init__()

        if structure == "hierarchical":
            structure = "hi"
        elif structure == "flattened":
            raise NotImplementedError("Flattened structure is not supported yet.")
        self.structure = structure

        self.pad_token_id = pad_token_id

        self.ignore_index = ignore_index

        self.apply_mask = apply_mask
        self.mask_prob = mask_prob
        self.vocab_size = vocab_size
        self.mask_token_id = mask_token_id
        self.mask_unit = mask_unit
        self.sep_token_id = sep_token_id

        self.simclr = simclr

        for k, v in kwargs.items():
            self.__setattr__(k, v)

        self.data = []
        self.subjects = []
        self.shard_ids = []
        self.sizes = []
        for i, manifest_path in enumerate(manifest_paths):
            with open(manifest_path, "r") as f:
                data_i_root = f.readline().strip()
                shard_ids = []
                for j, line in enumerate(f):
                    if debug and j >= 300:
                        break
                    items = line.strip().split("\t")
                    assert len(items) == 3, line
                    subject_id, num_events, shard_id = items
                    self.subjects.append((i, subject_id))
                    self.sizes.append(int(num_events))
                    shard_ids.append(int(shard_id))

                data_i = {}
                unique_shard_ids = np.unique(shard_ids)
                for shard_id in unique_shard_ids:
                    data_i[shard_id] = h5pickle.File(os.path.join(data_i_root, f"{shard_id}.h5"))["ehr"]
                self.data.append(data_i)
                self.shard_ids.extend(shard_ids)
        logger.info(f"loaded {len(self.subjects)} samples from {len(manifest_paths)} dataset(s)")

    def __len__(self):
        return len(self.subjects)


class HierarchicalMEDSDataset(MEDSDataset):
    def __init__(
        self,
        manifest_paths: List[str],
        max_events: int = 256,
        label: bool = False,
        tasks: List[str] = None,
        num_labels: List[int] = None,
        dummy_token_id: int = 101,
        **kwargs,
    ):
        kwargs.pop("structure", None)
        super().__init__(manifest_paths=manifest_paths, structure="hierarchical", **kwargs)

        self.max_events = max_events
        self.label = label
        self.tasks = tasks
        self.num_labels = num_labels
        self.dummy_token_id = dummy_token_id

    def mask(self, tokens: Union[np.ndarray, torch.Tensor], **kwargs):
        for i, event in enumerate(tokens):
            tokens[i], _ = super().mask(event, **kwargs)
        return tokens

    def collator(self, samples):
        samples = [s for s in samples if s["input_ids"] is not None]
        if len(samples) == 0:
            return {}

        if self.simclr:
            input_ids = sum(
                [
                    [s["input_ids"][: len(s["input_ids"]) // 2], s["input_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
            type_ids = sum(
                [
                    [s["type_ids"][: len(s["input_ids"]) // 2], s["type_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
            dpe_ids = sum(
                [
                    [s["dpe_ids"][: len(s["input_ids"]) // 2], s["dpe_ids"][len(s["input_ids"]) // 2 :]]
                    for s in samples
                ],
                [],
            )
        else:
            input_ids = [s["input_ids"] for s in samples]
            type_ids = [s["type_ids"] for s in samples]
            dpe_ids = [s["dpe_ids"] for s in samples]

        sizes = [s.size(0) for s in input_ids]
        target_size = self.max_events

        collated_input_ids = (
            input_ids[0].new_zeros((len(input_ids), target_size, len(input_ids[0][0]))).long()
        )
        collated_type_ids = type_ids[0].new_zeros((len(type_ids), target_size, len(type_ids[0][0]))).long()
        collated_dpe_ids = dpe_ids[0].new_zeros((len(dpe_ids), target_size, len(dpe_ids[0][0]))).long()
        for i, size in enumerate(sizes):
            diff = size - target_size
            if diff == 0:
                collated_input_ids[i] = input_ids[i]
                collated_type_ids[i] = type_ids[i]
                collated_dpe_ids[i] = dpe_ids[i]
            elif diff < 0:
                collated_input_ids[i] = torch.cat(
                    [
                        input_ids[i],
                        input_ids[i].new_zeros(-diff, len(input_ids[i][0])),
                    ],
                    dim=0,
                )
                # add dummy token to the start of each padded event as the event encoder can be
                # crushed when all the input tokens are pad tokens
                collated_input_ids[i][diff:, 0] = self.dummy_token_id
                collated_type_ids[i] = torch.cat(
                    [
                        type_ids[i],
                        type_ids[i].new_zeros(-diff, len(type_ids[i][0])),
                    ],
                    dim=0,
                )
                collated_dpe_ids[i] = torch.cat(
                    [
                        dpe_ids[i],
                        dpe_ids[i].new_zeros(-diff, len(dpe_ids[i][0])),
                    ],
                    dim=0,
                )
            else:
                collated_input_ids[i] = input_ids[i][-target_size:]
                collated_type_ids[i] = type_ids[i][-target_size:]
                collated_dpe_ids[i] = dpe_ids[i][-target_size:]

        out = {"id": [s["id"] for s in samples]}
        out["net_input"] = {
            "input_ids": collated_input_ids,
            "type_ids": collated_type_ids,
            "dpe_ids": collated_dpe_ids,
        }

        if self.label:
            label = {}
            for task in self.tasks:
                if len(samples[0][task]) == 1:
                    label[task] = torch.cat([s[task] for s in samples])
                else:
                    label[task] = torch.stack([s[task] for s in samples])
            out["label"] = label

        return out

    def __getitem__(self, idx):
        data_idx, subject = self.subjects[idx]
        data = self.data[data_idx][self.shard_ids[idx]][subject]

        tokens = data[self.structure][:]
        if self.apply_mask:
            tokens = self.mask(
                tokens,
                mask_prob=self.mask_prob,
                vocab_size=self.vocab_size,
                mask_token_id=self.mask_token_id,
                mask_unit=self.mask_unit,
                sep_token_id=self.sep_token_id,
            )

        ret = {
            "id": subject,
            "input_ids": torch.LongTensor(tokens[:, 0, :]),
            "type_ids": torch.LongTensor(tokens[:, 1, :]),
            "dpe_ids": torch.LongTensor(tokens[:, 2, :]),
        }

        if self.label:
            for i, task in enumerate(self.tasks):
                try:
                    ret[task] = torch.LongTensor(data["label"][i])
                except ValueError:
                    ret[task] = torch.LongTensor([data["label"][()]])
        return ret



import logging
import os
import pickle
import re
import shutil
import subprocess
import sys
from functools import reduce
from itertools import chain
from typing import List, Union

import h5py
import numpy as np
import pandas as pd
import pyspark.sql.functions as F
from pyspark.sql.types import ArrayType, IntegerType, StructField, StructType
from tqdm import tqdm
from transformers import AutoTokenizer
from utils import col_name_add, q_cut

logger = logging.getLogger(__name__)


class EHR(object):
    def __init__(self, cfg):
        self.cfg = cfg

        self.cache = cfg.cache

        cache_dir = os.path.expanduser("~/.cache/ehr")
        # cache_dir = self.cfg.dest
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        self.cache_dir = cache_dir

        if self.cache:
            logger.warning(
                "--cache is set to True. Note that it forces to load cached"
                f" data from {cache_dir},"
                " which may ignore some arguments such as --first_icu, as well as task-related"
                " arguments (--mortality, --los_3day, etc.)"
                " If you want to avoid this, do not set --cache to True."
            )

        self.data_dir = cfg.data
        self.ccs_path = cfg.ccs
        self.gem_path = cfg.gem
        self.ext = cfg.ext

        self.max_event_size = cfg.max_event_size if cfg.max_event_size is not None else sys.maxsize
        self.min_event_size = cfg.min_event_size if cfg.min_event_size is not None else 1
        assert self.min_event_size > 0, (
            "--min_event_size could not be negative or zero",
            self.min_event_size,
        )
        assert self.min_event_size <= self.max_event_size, (
            self.min_event_size,
            self.max_event_size,
        )

        self.max_event_token_len = cfg.max_event_token_len
        self.max_patient_token_len = cfg.max_patient_token_len

        self.max_age = cfg.max_age if cfg.max_age is not None else sys.maxsize
        self.min_age = cfg.min_age if cfg.min_age is not None else 0
        assert self.min_age <= self.max_age, (self.min_age, self.max_age)

        self.obs_size = cfg.obs_size
        self.gap_size = cfg.gap_size
        self.pred_size = cfg.pred_size
        self.long_term_pred_size = cfg.long_term_pred_size

        self.first_icu = cfg.first_icu

        # Emb_type / feature_select
        self.emb_type = cfg.emb_type
        self.feature = cfg.feature
        self.bucket_num = cfg.bucket_num

        # tasks
        self.mortality = cfg.mortality
        self.long_term_mortality = cfg.long_term_mortality
        self.los_3day = cfg.los_3day
        self.los_7day = cfg.los_7day
        self.readmission = cfg.readmission
        self.final_acuity = cfg.final_acuity
        self.imminent_discharge = cfg.imminent_discharge
        self.diagnosis = cfg.diagnosis
        self.creatinine = cfg.creatinine
        self.bilirubin = cfg.bilirubin
        self.platelets = cfg.platelets
        self.wbc = cfg.wbc

        self.chunk_size = cfg.chunk_size

        self.dest = cfg.dest

        self.bins = cfg.bins

        os.makedirs(os.path.join(self.dest, "data"), exist_ok=True)
        os.makedirs(os.path.join(self.dest, "metadata"), exist_ok=True)

        self.special_tokens_dict = dict()
        self.max_special_tokens = 100

        self.tokenizer = AutoTokenizer.from_pretrained("emilyalsentzer/Bio_ClinicalBERT")

        if self.emb_type == "textbase":
            self.cls_token_id = self.tokenizer.cls_token_id
            self.sep_token_id = self.tokenizer.sep_token_id

        elif self.emb_type == "codebase":
            self.cls_token_id = 1
            self.sep_token_id = 2

        self.table_type_id = 1
        self.column_type_id = 2
        self.value_type_id = 3
        self.timeint_type_id = 4
        self.cls_type_id = 5
        self.sep_type_id = 6

        self.others_dpe_id = 0

        self._icustay_fname = None
        self._patient_fname = None
        self._admission_fname = None
        self._diagnosis_fname = None

        self._icustay_key = None
        self._hadm_key = None

    @property
    def icustay_fname(self):
        return self._icustay_fname

    @property
    def patient_fname(self):
        return self._patient_fname

    @property
    def admission_fname(self):
        return self._admission_fname

    @property
    def diagnosis_fname(self):
        return self._diagnosis_fname

    @property
    def icustay_key(self):
        return self._icustay_key

    @property
    def hadm_key(self):
        return self._hadm_key

    @property
    def patient_key(self):
        return self._patient_key

    @property
    def determine_first_icu(self):
        return self._determine_first_icu

    @property
    def num_special_tokens(self):
        return len(self.special_tokens_dict)

    def build_cohorts(self, icustays, cached=False):
        if cached:
            cohorts = self.load_from_cache(self.ehr_name + ".cohorts")
            if cohorts is not None:
                return cohorts

        if not self.is_compatible(icustays):
            raise AssertionError(
                "{} do not have required columns to build cohorts.".format(self.icustay_fname)
                + " Please make sure that dataframe for icustays is compatible with other ehrs."
            )

        logger.info("Start building cohorts for {}".format(self.ehr_name))
        logger.info("Emb_type {}, Feature {}".format(self.emb_type, self.feature))
        obs_size = self.obs_size
        gap_size = self.gap_size

        icustays = icustays[icustays["LOS"] >= (obs_size + gap_size) / 24]
        icustays = icustays[(self.min_age <= icustays["AGE"]) & (icustays["AGE"] <= self.max_age)]

        # we define labels for the readmission task in this step
        # since it requires to observe each next icustays,
        # which would have been excluded in the final cohorts
        icustays.sort_values([self.hadm_key, self.icustay_key], inplace=True)

        if self.readmission:
            icustays["readmission"] = 1
            icustays.loc[
                icustays.groupby(self.hadm_key)[self.determine_first_icu].idxmax(), "readmission"
            ] = 0
        if self.first_icu:
            icustays = icustays.loc[icustays.groupby(self.hadm_key)[self.determine_first_icu].idxmin()]

        logger.info("cohorts have been built successfully. Loaded {} cohorts.".format(len(icustays)))

        self.save_to_cache(icustays, self.ehr_name + ".cohorts")

        return icustays

    # TODO process specific tasks according to user choice?
    def prepare_tasks(self, cohorts, spark, cached=False):
        if cached:
            labeled_cohorts = self.load_from_cache(self.ehr_name + ".cohorts.labeled")
            if labeled_cohorts is not None:
                return labeled_cohorts
            else:
                raise RuntimeError()

        logger.info("Start labeling cohorts for predictive tasks.")

        labeled_cohorts = cohorts[
            [
                self.hadm_key,
                self.icustay_key,
                self.patient_key,
                "readmission",
                "LOS",
                "INTIME",
                "OUTTIME",
                "DISCHTIME",
                "IN_ICU_MORTALITY",
                "HOS_DISCHARGE_LOCATION",
            ]
        ].copy()

        # mortality prediction
        # if the discharge location of an icustay is 'Death'
        #   & intime + obs_size + gap_size <= dischtime <= intime + obs_size + pred_size
        # it is assigned positive label on the mortality prediction

        if self.mortality:
            labeled_cohorts["mortality"] = (
                (
                    (labeled_cohorts["IN_ICU_MORTALITY"] == "Death")
                    | (labeled_cohorts["HOS_DISCHARGE_LOCATION"] == "Death")
                )
                & (self.obs_size * 60 + self.gap_size * 60 < labeled_cohorts["DISCHTIME"])
                & (labeled_cohorts["DISCHTIME"] <= self.obs_size * 60 + self.pred_size * 60)
            ).astype(int)

        if self.long_term_mortality:
            labeled_cohorts["long_term_mortality"] = (
                (
                    (labeled_cohorts["IN_ICU_MORTALITY"] == "Death")
                    | (labeled_cohorts["HOS_DISCHARGE_LOCATION"] == "Death")
                )
                & (self.obs_size * 60 + self.gap_size * 60 < labeled_cohorts["DISCHTIME"])
                & (labeled_cohorts["DISCHTIME"] <= self.obs_size * 60 + self.long_term_pred_size * 60)
            ).astype(int)

        if self.los_3day:
            labeled_cohorts["los_3day"] = (labeled_cohorts["LOS"] > 3).astype(int)
        if self.los_7day:
            labeled_cohorts["los_7day"] = (labeled_cohorts["LOS"] > 7).astype(int)

        if self.final_acuity or self.imminent_discharge:
            # if the discharge of 'Death' occurs in icu or hospital
            # we retain these cases for the imminent discharge task
            labeled_cohorts["IN_HOSPITAL_MORTALITY"] = (
                (~labeled_cohorts["IN_ICU_MORTALITY"])
                & (labeled_cohorts["HOS_DISCHARGE_LOCATION"] == "Death")
            ).astype(int)

            if self.final_acuity:
                # define final acuity prediction task
                labeled_cohorts["final_acuity"] = labeled_cohorts["HOS_DISCHARGE_LOCATION"]
                labeled_cohorts.loc[
                    labeled_cohorts["IN_ICU_MORTALITY"] == 1, "final_acuity"
                ] = "IN_ICU_MORTALITY"
                labeled_cohorts.loc[
                    labeled_cohorts["IN_HOSPITAL_MORTALITY"] == 1, "final_acuity"
                ] = "IN_HOSPITAL_MORTALITY"

                with open(os.path.join(self.dest, "metadata/final_acuity_classes.tsv"), "w") as f:
                    for i, cat in enumerate(
                        labeled_cohorts["final_acuity"].astype("category").cat.categories
                    ):
                        print("{}\t{}".format(i, cat), file=f)
                labeled_cohorts["final_acuity"] = labeled_cohorts["final_acuity"].astype("category").cat.codes
                # Replace -1 with NaN
                labeled_cohorts.loc[labeled_cohorts["final_acuity"] == -1, "final_acuity"] = np.nan
            if self.imminent_discharge:
                # define imminent discharge prediction task
                is_discharged = (self.obs_size * 60 + self.gap_size * 60 <= labeled_cohorts["DISCHTIME"]) & (
                    labeled_cohorts["DISCHTIME"] <= self.obs_size * 60 + self.pred_size * 60
                )
                labeled_cohorts.loc[is_discharged, "imminent_discharge"] = labeled_cohorts.loc[
                    is_discharged, "HOS_DISCHARGE_LOCATION"
                ]
                labeled_cohorts.loc[
                    is_discharged
                    & (
                        (labeled_cohorts["IN_ICU_MORTALITY"] == 1)
                        | (labeled_cohorts["IN_HOSPITAL_MORTALITY"] == 1)
                    ),
                    "imminent_discharge",
                ] = "Death"
                labeled_cohorts.loc[~is_discharged, "imminent_discharge"] = "No Discharge"

                with open(os.path.join(self.dest, "metadata/imminent_discharge_classes.tsv"), "w") as f:
                    for i, cat in enumerate(
                        labeled_cohorts["imminent_discharge"].astype("category").cat.categories
                    ):
                        print("{}\t{}".format(i, cat), file=f)
                labeled_cohorts["imminent_discharge"] = (
                    labeled_cohorts["imminent_discharge"].astype("category").cat.codes
                )
                # Replace -1 with NaN
                labeled_cohorts.loc[
                    labeled_cohorts["imminent_discharge"] == -1, "imminent_discharge"
                ] = np.nan
            labeled_cohorts = labeled_cohorts.drop(columns=["IN_HOSPITAL_MORTALITY"])

        # clean up unnecessary columns
        labeled_cohorts = labeled_cohorts.drop(
            columns=["LOS", "IN_ICU_MORTALITY", "DISCHTIME", "HOS_DISCHARGE_LOCATION"]
        )

        self.save_to_cache(labeled_cohorts, self.ehr_name + ".cohorts.labeled")

        logger.info("Done preparing tasks except for diagnosis prediction.")

        return labeled_cohorts

    def process_tables(self, cohorts, spark):
        # in: cohorts, sparksession
        # out: Spark DataFrame with (stay_id, time offset, inp, type, dpe)
        if isinstance(cohorts, pd.DataFrame):
            cohorts = cohorts[[self.hadm_key, self.icustay_key, "INTIME", "OUTTIME"]]
            logger.info("Start Preprocessing Tables, Cohort Numbers: {}".format(len(cohorts)))
            cohorts = spark.createDataFrame(cohorts)
            print("Converted Cohort to Pyspark DataFrame")
        else:
            logger.info("Start Preprocessing Tables")

        events_dfs = []
        for table_index, table in enumerate(self.tables):
            fname = table["fname"]
            table_name = fname.split("/")[-1][: -len(self.ext)]

            timestamp_key = table["timestamp"]
            excludes = table["exclude"]
            obs_size = self.obs_size
            logger.info("{} in progress.".format(fname))

            code_to_descriptions = None
            if "code" in table:
                code_to_descriptions = {
                    k: pd.read_csv(os.path.join(self.data_dir, v))
                    for k, v in zip(table["code"], table["desc"])
                }
                code_to_descriptions = {
                    k: dict(zip(v[k], v[d_k]))
                    for (k, v), d_k in zip(code_to_descriptions.items(), table["desc_key"])
                }

            infer_icustay_from_hadm_key = False

            events = spark.read.csv(os.path.join(self.data_dir, fname), header=True)
            if self.icustay_key not in events.columns:
                infer_icustay_from_hadm_key = True
                if self.hadm_key not in events.columns:
                    raise AssertionError(
                        "{} doesn't have one of these columns: {}".format(
                            fname, [self.icustay_key, self.hadm_key]
                        )
                    )

            events = events.drop(*excludes)
            if table["timeoffsetunit"] == "abs":
                events = events.withColumn(timestamp_key, F.to_timestamp(timestamp_key))

            if infer_icustay_from_hadm_key:
                events = events.join(
                    cohorts.select(self.hadm_key, self.icustay_key, "INTIME", "OUTTIME"),
                    on=self.hadm_key,
                    how="inner",
                )
                if table["timeoffsetunit"] == "abs":
                    events = (
                        events.withColumn(
                            "TEMP_TIME",
                            F.round((F.col(timestamp_key).cast("long") - F.col("INTIME").cast("long")) / 60),
                        )
                        .filter(F.col("TEMP_TIME") >= 0)
                        .filter(F.col("TEMP_TIME") <= F.col("OUTTIME"))
                        .drop("TEMP_TIME")
                    )
                else:
                    # All tables in eICU has icustay_key -> no need to handle
                    raise NotImplementedError()
                events = events.join(cohorts.select(self.icustay_key), on=self.icustay_key, how="leftsemi")

            else:
                events = events.join(
                    cohorts.select(self.icustay_key, "INTIME", "OUTTIME"), on=self.icustay_key, how="inner"
                )

            if table["timeoffsetunit"] == "abs":
                events = events.withColumn(
                    "TIME", F.round((F.col(timestamp_key).cast("long") - F.col("INTIME").cast("long")) / 60)
                )
                events = events.drop(timestamp_key)
            elif table["timeoffsetunit"] == "min":
                events = events.withColumn("TIME", F.col(timestamp_key).cast("int"))
                events = events.drop(timestamp_key)
            else:
                raise NotImplementedError()

            events = events.filter(F.col("TIME") >= 0).filter(F.col("TIME") <= obs_size * 60)
            events = events.drop("INTIME", "OUTTIME", self.hadm_key)

            if code_to_descriptions:
                for col in code_to_descriptions.keys():
                    mapping_expr = F.create_map([F.lit(x) for x in chain(*code_to_descriptions[col].items())])
                    events = events.withColumn(col, mapping_expr[F.col(col)])

            if self.emb_type == "codebase":
                print("codebase pre-process starts")
                events = events.toPandas()
                print("Converted events Pyspark DataFrame to Pandas DataFrame")

                code_col = table["code_feat"][0]
                dtype_code = events[code_col].dtype

                # Numericl feature bucketize
                for numeric_col in table["numeric_feat"]:
                    if numeric_col in events.columns:
                        # numeric / not_numeric classifying
                        numeric = events[pd.to_numeric(events[numeric_col], errors="coerce").notnull()]
                        numeric = numeric.astype({numeric_col: "float"})
                        not_numeric = events[pd.to_numeric(events[numeric_col], errors="coerce").isnull()]

                        # buckettize
                        numeric.loc[:, numeric_col] = numeric.groupby(code_col)[numeric_col].transform(
                            lambda x: x.rank(method="dense")
                        )
                        numeric.loc[:, numeric_col] = numeric.groupby(code_col)[numeric_col].transform(
                            lambda x: q_cut(x, self.bucket_num)
                        )

                        numeric[numeric_col] = (
                            "B_" + numeric[code_col].astype("str") + "_" + numeric[numeric_col].astype("str")
                        )

                        events = pd.concat([numeric, not_numeric], axis=0)

                # Categorical feature categorize
                for cate_col in table["categorical_feat"]:
                    if cate_col in events.columns:
                        events.loc[:, cate_col] = events[cate_col].map(lambda x: col_name_add(x, cate_col))

                events = events.astype({code_col: dtype_code})

                table_feature_unique = []
                for col in events.columns:
                    if col in [self.icustay_key, self.hadm_key, self.patient_key, "TIME"]:
                        continue
                    col_unique = list(events[col].unique())
                    table_feature_unique.extend(col_unique)

                table_feature_unique = list(set(table_feature_unique))

                if len(events_dfs) == 0:
                    max_idx = 3 + len(self.tables)
                    self.table_feature_dict = {k: idx + max_idx for idx, k in enumerate(table_feature_unique)}
                else:
                    max_idx = max(self.table_feature_dict.values())
                    table_feature_unique = [
                        k for k in table_feature_unique if k not in self.table_feature_dict.keys()
                    ]
                    self.table_feature_dict.update(
                        {k: idx + max_idx for idx, k in enumerate(table_feature_unique)}
                    )

                encoded_table_name = ([3 + table_index], [self.table_type_id], [0])
                encoded_cols = {
                    k: ([max(self.table_feature_dict.values()) + 1 + i], [self.column_type_id], [0])
                    for i, k in enumerate(events.columns)
                    if k not in [self.icustay_key, self.hadm_key, self.patient_key, "TIME"]
                }
                for i, k in enumerate(events.columns):
                    if k not in [self.icustay_key, self.hadm_key, self.patient_key, "TIME"]:
                        self.table_feature_dict[k] = max(self.table_feature_dict.values()) + 1 + i

                print("length of codebook = ", len(self.table_feature_dict))

                events = events.fillna(" ")
                events = spark.createDataFrame(events)
                print("Converted Events DataFrame to Pyspark DataFrame")

                def process_unit(feature, type_id):
                    input_ids = [self.table_feature_dict[feature]]
                    types = [type_id]
                    dpes = [0]
                    return input_ids, types, dpes

            elif self.emb_type == "textbase":
                print("textbase pre-process starts")

                def process_unit(text, type_id):
                    # Given (table_name|col|val), generate ([inp], [type], [dpe])
                    text = re.sub(r"\d*\.\d+", lambda x: str(round(float(x.group(0)), 4)), str(text))
                    number_groups = [g for g in re.finditer(r"([0-9]+([.][0-9]*)?|[0-9]+|\.+)", text)]
                    text = re.sub(r"([0-9\.])", r" \1 ", text)
                    input_ids = self.tokenizer.encode(text, add_special_tokens=False)
                    types = [type_id] * len(input_ids)

                    def get_dpe(tokens, number_groups):
                        number_ids = [121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 119]
                        numbers = [i for i, j in enumerate(tokens) if j in number_ids]
                        numbers_cnt = 0
                        data_dpe = [0] * len(tokens)
                        for group in number_groups:
                            if group[0] == "." * len(group[0]):
                                numbers_cnt += len(group[0])
                                continue

                            start = numbers[numbers_cnt]
                            end = numbers[numbers_cnt + len(group[0]) - 1] + 1
                            corresponding_numbers = tokens[start:end]
                            digits = [i for i, j in enumerate(corresponding_numbers) if j == 119]

                            # Case Integer
                            if len(digits) == 0:
                                data_dpe[start:end] = list(range(len(group[0]) + 5, 5, -1))
                            # Case Float
                            elif len(digits) == 1:
                                digit_idx = len(group[0]) - digits[0]
                                data_dpe[start:end] = list(
                                    range(len(group[0]) + 5 - digit_idx, 5 - digit_idx, -1)
                                )
                            else:
                                logger.warn(f"{data_dpe[start:end]} has irregular numerical formats")

                            numbers_cnt += len(group[0])
                        return data_dpe

                    dpes = get_dpe(input_ids, number_groups)
                    return input_ids, types, dpes

                encoded_table_name = process_unit(table_name, self.table_type_id)  # table name
                encoded_cols = {k: process_unit(k, self.column_type_id) for k in events.columns}  # table cols

            schema = StructType(
                [
                    StructField("INPUTS", ArrayType(IntegerType()), False),
                    StructField("TYPES", ArrayType(IntegerType()), False),
                    StructField("DPES", ArrayType(IntegerType()), False),
                ]
            )

            # encoded_table_name -> tuple(Input, type, dpe)
            # encoded_cols -> dict(col_name: input,type, dpe)
            def process_row(encoded_table_name, encoded_cols):
                def _process_row(row):
                    """
                    input: row (cols: icustay_id, timestamp, ...)
                    output: (input, type, dpe)
                    """
                    row = row.asDict()
                    # Should INITIALIZE with blank arrays to prevent corruption in Pyspark... Why??
                    input_ids, types, dpes = [], [], []
                    input_ids += encoded_table_name[0]
                    types += encoded_table_name[1]
                    dpes += encoded_table_name[2]
                    encoded_table_name
                    for col, val in row.items():
                        if col in [self.icustay_key, "TIME"] or val is None or val == " ":
                            continue
                        encoded_col = encoded_cols[col]
                        encoded_val = process_unit(val, self.value_type_id)

                        # -> values code화

                        if (
                            len(input_ids) + len(encoded_col[0]) + len(encoded_val[0]) + 2
                            <= self.max_event_token_len
                        ):
                            input_ids += encoded_col[0] + encoded_val[0]
                            types += encoded_col[1] + encoded_val[1]
                            dpes += encoded_col[2] + encoded_val[2]
                        else:
                            break
                    return input_ids, types, dpes

                return F.udf(_process_row, returnType=schema)

            events = (
                events.withColumn(
                    "tmp", process_row(encoded_table_name, encoded_cols)(F.struct(*events.columns))
                )
                .withColumn("INPUTS", F.col("tmp.INPUTS"))
                .withColumn("TYPES", F.col("tmp.TYPES"))
                .withColumn("DPES", F.col("tmp.DPES"))
                .select(self.icustay_key, "TIME", "INPUTS", "TYPES", "DPES")  # column 을 얘네만 남김
            )  # event process # -> INPUTS / TYPES / DPES /

            events_dfs.append(events)

        if self.emb_type == "codebase":
            os.makedirs(os.path.join(self.cache_dir, self.ehr_name), exist_ok=True)
            with open(
                os.path.join(self.cache_dir, self.ehr_name, f"codebase_code2idx_{self.feature}.pkl"), "wb"
            ) as f:
                pickle.dump(self.table_feature_dict, f)

        return reduce(lambda x, y: x.union(y), events_dfs)

    def make_input(self, cohorts, events, spark):
        @F.pandas_udf(returnType="TIME int", functionType=F.PandasUDFType.GROUPED_MAP)
        def _make_input(events):
            # Actually, this function does not have to return anything.
            # However, return something(TIME) is required to satisfy the PySpark requirements.
            df = events.sort_values("TIME")
            flatten_cut_idx = -1
            # Consider SEP
            flatten_lens = np.cumsum(df["INPUTS"].str.len() + 1).values
            event_length = len(df)

            if flatten_lens[-1] > self.max_patient_token_len - 1:
                # Consider CLS token at first of the flatten input
                flatten_cut_idx = np.searchsorted(
                    flatten_lens, flatten_lens[-1] - self.max_patient_token_len + 1
                )
                flatten_lens = (flatten_lens - flatten_lens[flatten_cut_idx])[flatten_cut_idx + 1 :]
                event_length = len(flatten_lens)

            # Event length should not be longer than max_event_size
            event_length = min(event_length, self.max_event_size)
            df = df.iloc[-event_length:]

            if len(df) <= self.min_event_size:
                return events["TIME"].to_frame()

            def make_hi(cls_id, sep_id, iterable):
                return [[cls_id] + list(i) + [sep_id] for i in iterable]

            def make_fl(cls_id, sep_id, iterable):
                return [cls_id] + list(chain(*[list(i) + [sep_id] for i in iterable]))

            hi_input = make_hi(self.cls_token_id, self.sep_token_id, df["INPUTS"])
            hi_type = make_hi(self.cls_type_id, self.sep_type_id, df["TYPES"])
            hi_dpe = make_hi(self.others_dpe_id, self.others_dpe_id, df["DPES"])

            fl_input = make_fl(self.cls_token_id, self.sep_token_id, df["INPUTS"])
            fl_type = make_fl(self.cls_type_id, self.sep_type_id, df["TYPES"])
            fl_dpe = make_fl(self.others_dpe_id, self.others_dpe_id, df["DPES"])

            assert len(hi_input) <= self.max_event_size, hi_input
            assert all([len(i) <= self.max_event_token_len for i in hi_input]), hi_input
            assert len(fl_input) <= self.max_patient_token_len, fl_input

            # Add padding to save as numpy array
            hi_input = np.array(
                [np.pad(i, (0, self.max_event_token_len - len(i)), mode="constant") for i in hi_input]
            )
            hi_type = np.array(
                [np.pad(i, (0, self.max_event_token_len - len(i)), mode="constant") for i in hi_type]
            )
            hi_dpe = np.array(
                [np.pad(i, (0, self.max_event_token_len - len(i)), mode="constant") for i in hi_dpe]
            )

            fl_input = np.pad(fl_input, (0, self.max_patient_token_len - len(fl_input)), mode="constant")
            fl_type = np.pad(fl_type, (0, self.max_patient_token_len - len(fl_type)), mode="constant")
            fl_dpe = np.pad(fl_dpe, (0, self.max_patient_token_len - len(fl_dpe)), mode="constant")

            stay_id = df[self.icustay_key].values[0]
            # Create caches (cannot write to hdf5 directly with pyspark)
            data = {
                "hi": np.stack([hi_input, hi_type, hi_dpe], axis=1).astype(np.int16),
                "fl": np.stack([fl_input, fl_type, fl_dpe], axis=0).astype(np.int16),
                "time": df["TIME"].values,
            }
            with open(
                os.path.join(self.cache_dir, self.ehr_name, self.emb_type, self.feature, f"{stay_id}.pkl"),
                "wb",
            ) as f:
                pickle.dump(data, f)
            return events["TIME"].to_frame()

        os.makedirs(os.path.join(self.cache_dir, self.ehr_name, self.emb_type, self.feature), exist_ok=True)

        events.groupBy(self.icustay_key).apply(_make_input).write.mode("overwrite").format("noop").save()

        logger.info("Finish Data Preprocessing. Start to write to hdf5")

        f = h5py.File(os.path.join(self.dest, "data/data.h5"), "w")
        ehr_g = f.create_group("ehr")

        active_stay_ids = []

        for stay_id_file in tqdm(
            os.listdir(os.path.join(self.cache_dir, self.ehr_name, self.emb_type, self.feature))
        ):  # 45080 / 45080
            stay_id = stay_id_file.split(".")[0]
            with open(
                os.path.join(self.cache_dir, self.ehr_name, self.emb_type, self.feature, stay_id_file), "rb"
            ) as f:
                data = pickle.load(f)
            stay_g = ehr_g.create_group(str(stay_id))
            stay_g.create_dataset("hi", data=data["hi"], dtype="i2", compression="lzf", shuffle=True)
            stay_g.create_dataset("fl", data=data["fl"], dtype="i2", compression="lzf", shuffle=True)
            stay_g.create_dataset("time", data=data["time"], dtype="i")
            active_stay_ids.append(int(stay_id))

        shutil.rmtree(
            os.path.join(self.cache_dir, self.ehr_name, self.emb_type, self.feature), ignore_errors=True
        )
        # Drop patients with few events

        if not isinstance(cohorts, pd.DataFrame):
            cohorts = cohorts.toPandas()
            print(cohorts)

        logger.info(
            "Total {} patients in the cohort are skipped due to few events".format(
                len(cohorts) - len(active_stay_ids)
            )
        )
        cohorts = cohorts[cohorts[self.icustay_key].isin(active_stay_ids)]

        # Should consider pat_id for split
        # for seed in self.seed:
        #     shuffled = cohorts.groupby(
        #         self.patient_key
        #     )[self.patient_key].count().sample(frac=1, random_state=seed)
        #     cum_len = shuffled.cumsum()

        #     cohorts.loc[cohorts[self.patient_key].isin(
        #         shuffled[cum_len < int(sum(shuffled)*self.valid_percent)].index), f'split_{seed}'] = 'test'
        #     cohorts.loc[cohorts[self.patient_key].isin(
        #         shuffled[(cum_len >= int(sum(shuffled)*self.valid_percent))
        #         & (cum_len < int(sum(shuffled)*2*self.valid_percent))].index), f'split_{seed}'] = 'valid'
        #     cohorts.loc[
        #         cohorts[self.patient_key].isin(
        #             shuffled[cum_len >= int(sum(shuffled)*2*self.valid_percent)].index
        #         ), f'split_{seed}'
        #     ] = 'train'

        # select columns stay_id and tasks
        task_columns = [
            "mortality",
            "long_term_mortality",
            "los_3day",
            "los_7day",
            "readmission",
            "final_acuity",
            "imminent_discharge",
        ]
        selected_columns = [self.icustay_key] + [col for col in task_columns if col in cohorts.columns]
        cohorts = cohorts[selected_columns]
        cohorts.rename(columns={self.icustay_key: "stay_id"}, inplace=True)

        cohorts.to_csv(os.path.join(self.dest, "data/label.csv"), index=False)

        # Record corhots df to hdf5
        for _, row in cohorts.iterrows():
            group = ehr_g[str(row["stay_id"])]
            for col in cohorts.columns:
                if isinstance(row[col], (pd.Timestamp, pd.Timedelta)):
                    continue
                group.attrs[col] = row[col]
        f.close()
        logger.info("Done encoding events.")

        return

    def run_pipeline(self, spark) -> None:
        cohorts = self.build_cohorts(cached=self.cache)
        labeled_cohorts = self.prepare_tasks(cohorts, spark, cached=self.cache)
        events = self.process_tables(labeled_cohorts, spark)
        self.make_input(labeled_cohorts, events, spark)

    def add_special_tokens(self, new_special_tokens: Union[str, List]) -> None:
        if isinstance(new_special_tokens, str):
            new_special_tokens = [new_special_tokens]

        num_special_tokens = self.num_special_tokens
        overlapped = []
        for new_special_token in new_special_tokens:
            if new_special_token in self.special_tokens_dict:
                overlapped.append(new_special_token)

        if len(overlapped) > 0:
            logger.warn(
                "There are some tokens that have already been set to special tokens."
                " Please provide only NEW tokens. Aborted."
            )
            return None
        elif num_special_tokens + len(new_special_tokens) > self.max_special_tokens:
            logger.warn(
                f"Total additional special tokens should be less than {self.max_special_tokens}" " Aborted."
            )
            return None

        self.special_tokens_dict.update(
            {
                k: "[unused{}]".format(i)
                for i, k in enumerate(new_special_tokens, start=num_special_tokens + 1)
            }
        )

    def make_compatible(self, icustays):
        """
        make different ehrs compatible with one another here
        NOTE: outtime/dischtime is converted to relative minutes from intime
            but, maintain the intime as the original value for later use
        """
        raise NotImplementedError()

    def is_compatible(self, icustays):
        checklist = [
            self.hadm_key,
            self.icustay_key,
            self.patient_key,
            "LOS",
            "AGE",
            "INTIME",
            "OUTTIME",
            "DISCHTIME",
            "IN_ICU_MORTALITY",
            "HOS_DISCHARGE_LOCATION",
        ]
        for item in checklist:
            if item not in icustays.columns.to_list():
                return False
        return True

    def save_to_cache(self, f, fname, use_pickle=False) -> None:
        if use_pickle:
            import pickle

            with open(os.path.join(self.cache_dir, fname), "wb") as fptr:
                pickle.dump(f, fptr)
        else:
            f.to_pickle(os.path.join(self.cache_dir, fname))

    def load_from_cache(self, fname):
        cached = os.path.join(self.cache_dir, fname)
        if os.path.exists(cached):
            data = pd.read_pickle(cached)

            logger.info("Loaded data from {}".format(cached))
            return data
        else:
            return None

    def infer_data_extension(self) -> str:
        raise NotImplementedError()

    def download_ehr_from_url(self, url, dest) -> None:
        username = input("Email or Username: ")
        subprocess.run(
            [
                "wget",
                "-r",
                "-N",
                "-c",
                "np",
                "--user",
                username,
                "--ask-password",
                url,
                "-P",
                dest,
            ]
        )
        output_dir = url.replace("https://", "").replace("http://", "")

        if not os.path.exists(os.path.join(dest, output_dir)):
            raise AssertionError(
                "Download failed. Please check your network connection or "
                "if you log in with a credentialed user"
            )

    def download_ccs_from_url(self, dest) -> None:
        subprocess.run(
            [
                "wget",
                "-N",
                "-c",
                "https://www.hcup-us.ahrq.gov/toolssoftware/ccs/Multi_Level_CCS_2015.zip",
                "-P",
                dest,
            ]
        )

        import zipfile

        with zipfile.ZipFile(os.path.join(dest, "Multi_Level_CCS_2015.zip"), "r") as zip_ref:
            zip_ref.extractall(os.path.join(dest, "foo.d"))
        os.rename(
            os.path.join(dest, "foo.d", "ccs_multi_dx_tool_2015.csv"),
            os.path.join(dest, "ccs_multi_dx_tool_2015.csv"),
        )
        os.remove(os.path.join(dest, "Multi_Level_CCS_2015.zip"))
        shutil.rmtree(os.path.join(dest, "foo.d"))

    def download_icdgem_from_url(self, dest) -> None:
        subprocess.run(
            [
                "wget",
                "-N",
                "-c",
                "https://data.nber.org/gem/icd10cmtoicd9gem.csv",
                "-P",
                dest,
            ]
        )



import argparse
import os

import h5py
import torch
from torch.nn.functional import one_hot
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

"""
Sample Dataset for Integrated-EHR-Pipeline
NOTE: The preprocessed token indices are stored as np.int16 type for efficienty,
        so overflow can be caused when vocab size > 32767.
 - ehr.cohort.labeled.index: dataframe pickle
    - index: should be resetted by `df.reset_index()`
    - [hi_start:hi_end] is the range of indices for the patient's history
    - split: one of ['train', 'val', 'test']
 - ehr.data: .np file with np.int16 type, (num_total_events, 3, max_word_len)
    - second dimension
        - 0: input_ids
        - 1: type_ids
        - 2: dpe_ids
"""


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ehr",
        type=str,
        required=True,
        choices=["mimiciii", "mimiciv", "eicu"],
        help="Name of the EHR dataset",
    )
    parser.add_argument("--data", type=str, required=True, help="Path to the preprocessed data")
    parser.add_argument(
        "--pred_target",
        type=str,
        required=True,
        choices=[
            "mortality",
            "readmission",
            "los_3day",
            "los_7day",
            "final_acuity",
            "imminent_discharge",
            "diagnosis",
        ],
        help="Prediction target",
    )
    parser.add_argument("--max_event_size", type=int, default=256, help="max event size to crop to")
    parser.add_argument(
        "--max_event_token_len", type=int, default=128, help="max token length for each event (Hierarchical)"
    )

    parser.add_argument(
        "--max_patient_token_len", type=int, default=8192, help="max token length for each patient (Flatten)"
    )
    return parser


class BaseEHRDataset(Dataset):
    def __init__(self, args, split):
        super().__init__()

        self.args = args
        self.data_path = os.path.join(args.data, f"{args.ehr}.h5")

        self.data = h5py.File(self.data_path, "r")["ehr"]
        self.keys = []
        for key in self.data.keys():
            if self.data[key].attrs["split"] == split:
                self.keys.append(key)
        self.pred_target = args.pred_target

        self.num_classes = {
            "mimiciii": {
                "mortality": 2,
                "readmission": 2,
                "los_3day": 2,
                "los_7day": 2,
                "final_acuity": 18,
                "imminent_discharge": 18,
                "diagnosis": 18,
            },
            "mimiciv": {
                "mortality": 2,
                "readmission": 2,
                "los_3day": 2,
                "los_7day": 2,
                "final_acuity": 14,
                "imminent_discharge": 14,
                "diagnosis": 18,
            },
            "eicu": {
                "mortality": 2,
                "readmission": 2,
                "los_3day": 2,
                "los_7day": 2,
                "final_acuity": 9,
                "imminent_discharge": 9,
                "diagnosis": 18,
            },
        }[self.args.ehr][self.args.pred_target]

    def __len__(self):
        return len(self.keys)

    def __getitem__(self, idx):
        raise NotImplementedError()

    def collate_fn(self, out):
        ret = dict()
        if len(out) == 1:
            for k, v in out[0].items():
                if k == "label":
                    ret[k] = (
                        one_hot(torch.LongTensor([v]), self.num_classes)
                        .float()
                        .reshape((1, self.num_classes))
                    )
                else:
                    ret[k] = v.unsqueeze(0)
        else:
            for k, v in out[0].items():
                if k == "label":
                    ret[k] = one_hot(torch.LongTensor([i[k] for i in out]), self.num_classes).float()
                else:
                    ret[k] = pad_sequence([i[k] for i in out], batch_first=True)
        return ret


class HierarchicalEHRDataset(BaseEHRDataset):
    def __init__(self, args, split):
        super().__init__(args, split)

    def __getitem__(self, idx):
        # NOTE: Warning occurs when converting np.int16 read-only array into tensor, but ignoreable

        data = self.data[self.keys[idx]]["hi"]
        label = self.data[self.keys[idx]].attrs[self.pred_target]
        return {
            "input_ids": torch.IntTensor(data[:, 0, :]),
            "type_ids": torch.IntTensor(data[:, 1, :]),
            "dpe_ids": torch.IntTensor(data[:, 2, :]),
            "label": label,
        }


class FlattenEHRDataset(BaseEHRDataset):
    def __init__(self, args, split):
        super().__init__(args, split)
        self.data_path = os.path.join(args.data, f"{args.ehr}.flat.npy")

    def __getitem__(self, idx):
        # NOTE: Warning occurs when converting np.int16 read-only array into tensor, but ignoreable
        data = self.data[self.keys[idx]]["fl"]
        label = self.data[self.keys[idx]].attrs[self.pred_target]
        return {
            "input_ids": torch.IntTensor(data[0, :]),
            "type_ids": torch.IntTensor(data[1, :]),
            "dpe_ids": torch.IntTensor(data[2, :]),
            "label": label,
        }


def main():
    args = get_parser().parse_args()
    dataset = HierarchicalEHRDataset(args, "train")
    print(dataset.__getitem__(1))
    dataset = FlattenEHRDataset(args, "train")
    print(dataset.__getitem__(1))
    pass


if __name__ == "__main__":
    main()



import math
from dataclasses import dataclass, field
from typing import Any, Dict, List, Tuple

import numpy as np
import torch
import torch.nn.functional as F

import genhpf.utils.utils as utils
from genhpf.criterions import BaseCriterion, register_criterion
from genhpf.criterions.criterion import CriterionConfig
from genhpf.loggings import meters, metrics
from genhpf.loggings.meters import safe_round


@dataclass
class BinaryCrossEntropyWithLogitsConfig(CriterionConfig):
    threshold: float = field(default=0.5, metadata={"help": "threshold value for binary classification"})


@register_criterion("binary_cross_entropy_with_logits", dataclass=BinaryCrossEntropyWithLogitsConfig)
class BinaryCrossEntropyWithLogits(BaseCriterion):
    def __init__(self, cfg: BinaryCrossEntropyWithLogitsConfig):
        super().__init__(cfg)

        if self.task_names is not None and len(self.task_names) > 1:
            raise ValueError(
                "binary_cross_entropy_with_logits only supports single task training."
                " if you want to train multiple tasks, use multi_task_criterion instead."
            )

        self.threshold = cfg.threshold

    def compute_loss(
        self, logits: torch.Tensor, targets: torch.Tensor, sample=None, net_output=None, model=None
    ) -> Tuple[torch.Tensor, List[float]]:
        assert (
            logits.size() == targets.size()
        ), f"logits and targets must have the same size: {logits.size()} vs {targets.size()}"
        targets = targets.float()
        loss = F.binary_cross_entropy_with_logits(input=logits, target=targets, reduction="sum")
        return loss, [loss.detach().item()]

    def get_sample_size(self, sample, targets: torch.Tensor) -> int:
        if "sample_size" in sample:
            sample_size = sample["sample_size"]
        else:
            sample_size = targets.numel()
        return sample_size

    def get_logging_outputs(
        self, logging_output, logits: torch.Tensor, targets: torch.Tensor, sample=None
    ) -> List[Dict[str, Any]]:
        with torch.no_grad():
            probs = torch.sigmoid(logits)
            outputs = probs > self.threshold

            if probs.numel() == 0:
                corr = 0
                count = 0
            else:
                count = float(probs.numel())
                corr = (outputs == targets).sum().item()

            logging_output["correct"] = corr
            logging_output["count"] = count

            # report aucs only in eval mode
            if not self.training:
                logging_output["_y_true"] = targets.cpu().numpy()
                logging_output["_y_score"] = probs.cpu().numpy()

        return logging_output

    @staticmethod
    def reduce_metrics(logging_outputs: List[Dict[str, Any]], prefix: str = None) -> None:
        """Aggregate logging outputs from data parallel training."""
        if prefix is None:
            prefix = ""
        elif prefix is not None and not prefix.endswith("_"):
            prefix = prefix + "_"

        loss_sum = utils.item(sum(log.get("loss", 0) for log in logging_outputs))

        sample_size = utils.item(sum(log.get("sample_size", 0) for log in logging_outputs))

        metrics.log_scalar(f"{prefix}loss", loss_sum / (sample_size or 1) / math.log(2), sample_size, round=3)

        if "_y_true" in logging_outputs[0] and "_y_score" in logging_outputs[0]:
            y_true = np.concatenate([log.get("_y_true", []) for log in logging_outputs])
            y_score = np.concatenate([log.get("_y_score", []) for log in logging_outputs])

            metrics.log_custom(meters.AUCMeter, f"_{prefix}auc", y_score, y_true)

        correct = sum(log.get("correct", 0) for log in logging_outputs)
        metrics.log_scalar(f"_{prefix}correct", correct)

        total = sum(log.get("count", 0) for log in logging_outputs)
        metrics.log_scalar(f"_{prefix}total", total)

        if total > 0:
            metrics.log_derived(
                f"{prefix}accuracy",
                lambda meters: safe_round(meters[f"_{prefix}correct"].sum / meters[f"_{prefix}total"].sum, 5)
                if meters[f"_{prefix}total"].sum > 0
                else float("nan"),
            )

    def post_validate(self, stats, agg, **kwargs):
        for key in agg.keys():
            if key.startswith("_") and key.endswith("auc"):
                stats[key[1:-3] + "auroc"] = agg[key].auroc
                stats[key[1:-3] + "auprc"] = agg[key].auprc

        return stats
